__version__ = '0.1.0'

from . import data, models, layers, criterions, launcher, metrics, utils
from freerec.dict2obj import Config


def check_version(version: str):
    if version != __version__:
        print(f"\033[1;31m [Warning] FreeRec version of {version} is required but current version is {__version__} \033[0m")




from typing import Optional

import torchdata.datapipes as dp

from .base import RecDataSet
from ..fields import SparseField
from ..tags import FEATURE, TARGET, ITEM
from ...dict2obj import Config


__all__ = ['Avazu_x1']


class Avazu_x1(RecDataSet):
    """A Kaggle challenge dataset for Avazu CTR prediction.
    |     Dataset      |   Total    |   #Train   | #Validation |   #Test   |
    | :--------------: | :--------: | :--------: | :---------: | :-------: |
    | Avazu_x1 (7:1:2) | 40,428,967 | 28,300,276 |  4,042,897  | 8,085,794 |

    This dataset contains about 10 days of labeled click-through data on mobile advertisements. 
    It has 22 feature fields including user features and advertisement attributes. 
    See [here](https://github.com/openbenchmark/BARS/tree/master/ctr_prediction/datasets/Avazu) for details.

    Attributes:
    ---

    _cfg: Config
        - target: SparseField
            Label
        - features: SparseField
            Hour + C1 - C21, in total of 22 features. 
    open_kw: Config
        - mode: 'rt'
        - delimiter: ','
        - skip_lines: 1
    """

    URL = "https://zenodo.org/record/5700987/files/Avazu_x1.zip"

    _cfg = Config(
        target = [SparseField('Label', na_value=None, dtype=int, transformer='none', tags=TARGET)],
        features = [SparseField(f"Feat{k}", na_value=-1, dtype=int, tags=[ITEM, FEATURE]) for k in range(1, 23)]
    )

    _cfg.fields = _cfg.target + _cfg.features

    open_kw = Config(mode='rt', delimiter=',', skip_lines=1)

    def file_filter(self, filename: str):
        return self.mode in filename

    def raw2data(self) -> dp.iter.IterableWrapper:
        datapipe = dp.iter.FileLister(self.path)
        datapipe = datapipe.filter(filter_fn=self.file_filter)
        datapipe = datapipe.open_files(mode=self.open_kw.mode)
        datapipe = datapipe.parse_csv(delimiter=self.open_kw.delimiter, skip_lines=self.open_kw.skip_lines)
        datapipe = datapipe.map(self.row_processer)
        # data = list(datapipe)
        # if self.mode == 'train':
        #     random.shuffle(data)
        # datapipe = dp.iter.IterableWrapper(data)
        return datapipe







from typing import Callable, Iterable, Tuple, Union, Dict, List

import torch
import numpy as np
from functools import partial, lru_cache

from .utils import safe_cast
from .preprocessing import X2X, Label2Index, Binarizer, MinMaxScaler, StandardScaler
from .tags import FieldTags, SPARSE, DENSE
from ..utils import errorLogger, warnLogger


__all__ = ['Field', 'DenseField', 'SparseField', 'Token', 'SparseToken', 'DenseToken', 'Tokenizer']

TRANSFORM = {
    "none": X2X,
    'label2index': Label2Index,
    'minmax': MinMaxScaler,
    'standard': StandardScaler,
}


class Field:
    """Fielding data by column.
    
    Attributes:
    ---

    name: str
        Name of this field.
    tags: set
        A set of FieldTags.
    dtype: torch.long|torch.float32
    caster: Callable
        Convert elements to specified dtype (int, float or str) with na_value.
    transformer: Callable
        Transformation.
    """
    
    def __init__(
        self, name: str, na_value: Union[str, int, float], dtype: Callable,
        transformer: Union[str, Callable] = 'none', tags: Union[FieldTags, Iterable[FieldTags]] = tuple()
    ):
        """
        Parameters:
        ---

        name: str 
            The name of the field.
        na_value: str, int or float
            Fill 'na' with na_value.
        dtype: str|int|float for safe_cast
        transformer: 'none'|'label2index'|'binary'|'minmax'|'standard'
        tags: Union[FieldTags, Iterable[FieldTags]]
            For quick retrieve.

        Examples:
        ---

        >>> from freerec.data.tags import USER, ITEM, ID, TARGET
        >>> User = SparseField('User', -1, int, tags=(USER, ID))
        >>> Item = SparseField('Item', -1, int, tags=(ITEM, ID))
        >>> Target = SparseField('Label', -1, int, transformer='none', tags=TARGET)
        """

        self.__name = name
        self.__na_value = na_value
        self.__tags = set()
        self.dtype = dtype
        self.caster = partial(safe_cast, dest_type=dtype, default=na_value)
        self.transformer = TRANSFORM[transformer]() if isinstance(transformer, str) else transformer
        if isinstance(tags, FieldTags):
            self.add_tag(tags)
        else:
            self.add_tag(*tags)

    def add_tag(self, *tags: FieldTags):
        """Add some tags.

        Parameters:
        ---
        *tags: FieldTags

        Examples:
        ---

        >>> from freerec.data.tags import ID
        >>> User = SparseField('User', -1, int)
        >>> User.add_tag(ID)

        """
        for tag in tags:
            if not isinstance(tag, FieldTags):
                warnLogger(f"FieldTags is expected but {type(tags)} received ...")
            self.__tags.add(tag)

    @property
    def tags(self):
        return self.__tags

    def match(self, *tags: FieldTags):
        """If current field matches the given tags, return True.
        
        Parameters:
        ---
        *tags: FieldTags


        Returns:
        ---

        - `True`: If all given tags are matched.
        - 'False': If any tag is not matched.

        Examples:
        ---

        >>> from freerec.data.tags import ID
        >>> User = SparseField('User', -1, int)
        >>> User.add_tag(ID)
        >>> User.match(ID)
        True
        >>> User.match(ID, Feature)
        False
        >>> User.match(Feature)
        False

        """
        return all(tag in self.__tags for tag in tags)

    def partial_fit(self, x: np.array) -> np.array:
        """Partially fit some data."""
        self.transformer.partial_fit(x)

    def transform(self, x: np.array) -> np.array:
        """Transform data."""
        return self.transformer.transform(x)

    @property
    def name(self):
        return self.__name

    @property
    def na_value(self):
        return self.__na_value
    
    @property
    def dtype(self):
        return self.__dtype

    @dtype.setter
    def dtype(self, val):
        if val in (int, str):
            self.__dtype = torch.long
        elif val == float:
            self.__dtype = torch.float32
        else:
            self.__dtype = val

    def __str__(self) -> str:
        tags = ','.join(map(str, self.tags))
        return f"{self.name}: [dtype: {self.dtype}, na_value: {self.na_value}, tags: {tags}]"


class SparseField(Field):

    def __init__(
        self, name: str, na_value: Union[str, int, float], dtype: Callable, 
        transformer: Union[str, Callable] = 'label2index', 
        tags: Union[FieldTags, Iterable[FieldTags]] = tuple()
    ):
        super().__init__(name, na_value, dtype, transformer, tags)
        self.add_tag(SPARSE)

    @property
    def count(self):
        """Return the number of classes."""
        if isinstance(self.transformer, Label2Index):
            return len(self.transformer.classes_)
        else:
            return None


class DenseField(Field):

    def __init__(
        self, name: str, na_value: Union[str, int, float], dtype: Callable, 
        transformer: Union[str, Callable] = 'minmax', 
        tags: Union[FieldTags, Iterable[FieldTags]] = tuple()
    ):
        super().__init__(name, na_value, dtype, transformer, tags)
        self.add_tag(DENSE)
   

class Fielder(tuple):


    @lru_cache(maxsize=4)
    def groupby(self, *tags: FieldTags) -> 'Fielder':
        """Return those fields matching given tags."""
        return Fielder(field for field in self if field.match(*tags))

    @lru_cache(maxsize=4)
    def groupbynot(self, *tags: FieldTags) -> 'Fielder':
        """Return those fields not matching given tags."""
        return Fielder(field for field in self if not field.match(*tags))

    def state_dict(self) -> Dict:
        """Return state dict of fields."""
        return {field.name: field.transformer for field in self}

    def load_state_dict(self, state_dict: Dict, strict: bool = False):
        for field in self:
            field.transformer = state_dict.get(field.name, field.transformer)

    def copy(self) -> 'Fielder':
        return Fielder(self)

    def __getitem__(self, index: Union[int, FieldTags, Iterable[FieldTags]]) -> Union[Field, 'Fielder', None]:
        """Get tokens by index.

        Parameters:
        ---

        index: Union[int, FieldTags, Iterable[FieldTags]]
            - `int`: Return the token at position `int`.
            - `FieldTags`: Return the tokens matching `FieldTags`.
            - `Iterable[FieldTags]`: Return the tokens matching `Iterable[FieldTags]`.
        
        Returns:
        ---

        - `Token`: If only one token matches given tags.
        - `None`: If none of tokens matches given tags.
        - `Tokenizer`: If more than one tokens match given tags.

        Examples:
        ---

        >>> from freerec.data.tags import USER
        >>> tokenizer[0];
        >>> user = tokenizer[USER]
        >>> isinstance(user, Token)
        True
        >>> isinstance(tokenizer[ID], Tokenizer)
        True
        """
        if isinstance(index, int):
            return super().__getitem__(index)
        elif isinstance(index, FieldTags):
            fields =  self.groupby(index)
        else:
            fields = self.groupby(*index)
        if len(fields) == 1:
            return fields[0]
        elif len(fields) == 0:
            return None # for a safety purpose
        else:
            return fields


class Token(torch.nn.Module):

    def __init__(self, field: Field) -> None:
        super().__init__()
        self.__name = field.name
        self.__tags = field.tags.copy()

        self.dimension: int
        self.embeddings: Union[torch.nn.Embedding, torch.nn.Linear, torch.nn.Identity]

    @property
    def name(self):
        return self.__name

    @property
    def tags(self):
        return self.__tags

    def add_tag(self, *tags: FieldTags):
        """Add some tags.

        Parameters:
        ---
        *tags: FieldTags

        Examples:
        ---

        >>> from freerec.data.tags import ID
        >>> User = SparseField('User', -1, int)
        >>> User.add_tag(ID)

        """
        for tag in tags:
            if not isinstance(tag, FieldTags):
                warnLogger(f"FieldTags is expected but {type(tags)} received ...")
            self.__tags.add(tag)

    def match(self, *tags: FieldTags):
        """If current field matches the given tags, return True.
        
        Parameters:
        ---
        *tags: FieldTags


        Returns:
        ---

        - `True`: If all given tags are matched.
        - 'False': If any tag is not matched.

        Examples:
        ---

        >>> from freerec.data.tags import ID
        >>> User = SparseField('User', -1, int)
        >>> User.add_tag(ID)
        >>> User.match(ID)
        True
        >>> User.match(ID, Feature)
        False
        >>> User.match(Feature)
        False

        """
        return all(tag in self.__tags for tag in tags)

    def embed(self, dim: int, **kwargs):
        errorLogger("embed method should be specified ...", NotImplementedError)


class SparseToken(Token):

    def __init__(self, field: Field) -> None:
        super().__init__(field)

        self.__count = field.count

    @property
    def count(self):
        """Return the number of classes."""
        return self.__count

    def embed(self, dim: int, **kwargs):
        """Create nn.Embedding.

        Parameters:
        ---

        dim: int
            Dimension.
        **kwargs: other kwargs for nn.Embedding

        Examples:
        ---

        >>> token = SparseToken(field)
        >>> token.embed(8)
        """

        self.dimension = dim
        self.embeddings = torch.nn.Embedding(self.count, dim, **kwargs)

    def look_up(self, x: torch.Tensor) -> torch.Tensor:
        """Look up embeddings for categorical features.

        Parameters:
        ---

        x: (B, *), torch.Tensor

        Returns:
        ---

        embeddings: (B, *, d)

        Examples:
        ---

        >>> User: SparseToken
        >>> ids = torch.arange(3).view(-1, 1)
        >>> User.look_up(ids).ndim
        3
        """
        return self.embeddings(x)


class DenseToken(Token):

    def embed(self, dim: int = 1, bias: bool = False, linear: bool = False):
        """Create Embedding in nn.Linear manner.

        Parameters:
        ---

        dim: int
            Dimension.
        bias: bool
            Add bias or not.
        linear: bool
            - `True`: nn.Linear.
            - `False`: using nn.Identity instead.

        Examples:
        ---

        >>> token = DenseToken(field)
        >>> token.embed(8, bias=True, linear=False)
        """
        if linear:
            self.dimension = dim
            self.embeddings = torch.nn.Linear(1, dim, bias=bias)
        else:
            self.dimension = 1
            self.embeddings = torch.nn.Identity()

    def look_up(self, x: torch.Tensor) -> torch.Tensor:
        """Look up embeddings for numeric features.
        Note that the return embeddings' shape is in accordance with
        that of categorical features.

        Parameters:
        ---

        x: (B, *), torch.Tensor

        Returns:
        ---

        embeddings: (B, *, 1) or (B, *, d)
            - If `linear` is True, it returns (B, *, d).
            - If `linear` is False, it returns (B, *, 1).

        >>> Feat: DenseToken
        >>> vals = torch.rand(3, 1)
        >>> Feat.look_up(vals).ndim
        3
        """
        return self.embeddings(x.unsqueeze(-1))


class Tokenizer(torch.nn.Module):
    """A collection of tokens.
    
    Attributes:
    ---

    tokens: nn.ModuleList

    """

    def __init__(self, fields: Iterable[Union[Field, Token]]) -> None:
        """
        Examples:
        ---

        >>> from freerec.data.datasets import Gowalla_m1
        >>> basepipe = Gowalla_m1("../data")
        >>> fields = basepipe.fields
        >>> tokenizer = Tokenizer(fields)
        """
        super().__init__()

        fields = [self.totoken(field) for field in fields]
        self.tokens = torch.nn.ModuleList(fields)
    
    def totoken(self, field: Field):
        """Field to Token."""
        if isinstance(field, Token):
            return field
        elif isinstance(field, SparseField):
            return SparseToken(field)
        elif isinstance(field, DenseField):
            return DenseToken(field)
        else:
            errorLogger("Only Sparse|DenseField supported !")

    @lru_cache(maxsize=4)
    def groupby(self, *tags: FieldTags) -> List[Token]:
        """Group by given tags and return list of tokens matched.

        Examples:
        ---
        >>> from freerec.data.tags import USER, ID
        >>> User = tokenizer.groupby(USER, ID)
        >>> isinstance(User, List)
        True
        >>> Item = tokenizer.groupby(ITEM, ID)[0]
        >>> Item.match(ITEM)
        True
        >>> Item.match(ID)
        True
        >>> Item.match(User)
        False
        """
        return [token for token in self.tokens if token.match(*tags)]

    def embed(self, dim: int, *tags: FieldTags, **kwargs):
        """Create embeddings.

        Parameters:
        ---
        dim: int
            Dimension.
        *tags: FieldTags
        **kwargs: kwargs for nn.Embeddings or nn.Linear

        Examples:
        ---

        >>> from freerec.data.tags import DENSE, SPARSE
        >>> tokenizer.embed(8, SPARSE) # nn.Embedding(count, 8)
        >>> tokenizer.embed(8, DENSE, bias=False) # nn.Linear(1, 8, bias=False)
        """
        for feature in self.groupby(*tags):
            feature.embed(dim, **kwargs)

    def look_up(self, inputs: Dict[str, torch.Tensor], *tags: FieldTags) -> List[torch.Tensor]:
        """Dict[str, torch.Tensor: B x K] -> List[torch.Tensor: B x K x d]
        
        Parameters:
        ---

        inputs: Dict[str, torch.Tensor]
            A dict of inputs.

        Returns:
        ---

        A list of torch.Tensor.

        Examples:
        ---

        >>> from freerec.data.tags import User, ID
        >>> userEmbds = tokenizer.look_up(users, USER, ID)
        >>> isinstance(userEmbds, List)
        True

        """
        return [token.look_up(inputs[token.name]) for token in self.groupby(*tags)]

    def calculate_dimension(self, *tags: FieldTags):
        """Return the total dimension of tokens matching tiven tags."""
        return sum(feature.dimension for feature in self.groupby(*tags))

    def __len__(self):
        return len(self.tokens)

    def __getitem__(self, index: Union[int, FieldTags, Iterable[FieldTags]]) -> Union[Token, 'Tokenizer', None]:
        """Get tokens by index.

        Parameters:
        ---

        index: Union[int, FieldTags, Iterable[FieldTags]]
            - `int`: Return the token at position `int`.
            - `FieldTags`: Return the tokens matching `FieldTags`.
            - `Iterable[FieldTags]`: Return the tokens matching `Iterable[FieldTags]`.
        
        Returns:
        ---

        - `Token`: If only one token matches given tags.
        - `None`: If none of tokens matches given tags.
        - `Tokenizer`: If more than one tokens match given tags.

        Examples:
        ---

        >>> from freerec.data.tags import USER
        >>> tokenizer[0];
        >>> user = tokenizer[USER]
        >>> isinstance(user, Token)
        True
        >>> isinstance(tokenizer[ID], Tokenizer)
        True
        """
        if isinstance(index, int):
            tokens = [self.tokens[index]]
        elif isinstance(index, FieldTags):
            tokens =  self.groupby(index)
        else:
            tokens = self.groupby(*index)
        if len(tokens) == 1:
            return tokens[0]
        elif len(tokens) == 0:
            return None # for a safety purpose
        else:
            return Tokenizer(tokens)





from typing import Optional, Tuple

import torch
import torchdata.datapipes as dp
import scipy.sparse as sp
import pandas as pd


def get_csr_matrix(
    datapipe: dp.iter.IterDataPipe, 
    rowName: str, colName: str, valName: str,
    shape: Optional[Tuple[int, int]] = None
) -> sp.csr_array:
    """Compressed Sparse Row matrix
    Advantages of the CSR format
      - efficient arithmetic operations CSR + CSR, CSR * CSR, etc.
      - efficient row slicing
      - fast matrix vector products

    Disadvantages of the CSR format
      - slow column slicing operations (consider CSC)
      - changes to the sparsity structure are expensive (consider LIL or DOK)
    """
    df = None
    for chunk in datapipe:
        df = pd.concat((df, chunk))
    df = df[[rowName, colName, valName]]
    df = df[df[valName] > 0]
    df = df.sort_values(by=[rowName, colName])
    rows, cols, vals = df[rowName].values, df[colName].values, df[valName].values
    return sp.csr_array((vals, (rows, cols)), shape=shape)

def get_csc_matrix(
    datapipe: dp.iter.IterDataPipe, 
    rowName: str, colName: str, valName: str,
    shape: Optional[Tuple[int, int]] = None
) -> sp.csc_array:
    """Compressed Sparse Column matrix
    Advantages of the CSC format
        - efficient arithmetic operations CSC + CSC, CSC * CSC, etc.
        - efficient column slicing
        - fast matrix vector products (CSR, BSR may be faster)

    Disadvantages of the CSC format
      - slow row slicing operations (consider CSR)
      - changes to the sparsity structure are expensive (consider LIL or DOK)
    """
    matrix = get_csr_matrix(datapipe, rowName, colName, valName, shape)
    return matrix.tocsc()

def get_coo_matrix(
    datapipe: dp.iter.IterDataPipe, 
    rowName: str, colName: str, valName: str,
    shape: Optional[Tuple[int, int]] = None
) -> sp.coo_array:
    """A sparse matrix in COOrdinate format.
    Also known as the 'ijv' or 'triplet' format.
    Advantages of the COO format
        - facilitates fast conversion among sparse formats
        - permits duplicate entries
        - very fast conversion to and from CSR/CSC formats

    Disadvantages of the COO format
        - does not directly support:
            + arithmetic operations
            + slicing
    """
    matrix = get_csr_matrix(datapipe, rowName, colName, valName, shape)
    return matrix.tocoo()

def get_lil_matrix(
    datapipe: dp.iter.IterDataPipe, 
    rowName: str, colName: str, valName: str,
    shape: Optional[Tuple[int, int]] = None
) -> sp.lil_array:
    """Row-based list of lists sparse matrix
    Advantages of the LIL format
        - supports flexible slicing
        - changes to the matrix sparsity structure are efficient

    Disadvantages of the LIL format
        - arithmetic operations LIL + LIL are slow (consider CSR or CSC)
        - slow column slicing (consider CSC)
        - slow matrix vector products (consider CSR or CSC)
    """
    matrix = get_csr_matrix(datapipe, rowName, colName, valName, shape)
    return matrix.tolil()

def get_dok_matrix(
    datapipe: dp.iter.IterDataPipe, 
    rowName: str, colName: str, valName: str,
    shape: Optional[Tuple[int, int]] = None
) -> sp.dok_array:
    """Dictionary Of Keys based sparse matrix.
    This is an efficient structure for constructing sparse
    matrices incrementally.
    """
    matrix = get_csr_matrix(datapipe, rowName, colName, valName, shape)
    return matrix.todok()

def sparse_matrix_to_coo_tensor(sparse_matrix):
    sparse_matrix = sparse_matrix.tocoo()
    rows = torch.from_numpy(sparse_matrix.row).long()
    cols = torch.from_numpy(sparse_matrix.col).long()
    vals = torch.from_numpy(sparse_matrix.data)
    indices = torch.stack((rows, cols), axis=0)
    return torch.sparse_coo_tensor(indices, vals, torch.Size(sparse_matrix.shape))

def sparse_matrix_to_csr_tensor(sparse_matrix):
    return sparse_matrix_to_coo_tensor(sparse_matrix).to_sparse_csr()




from .base import ImplicitRecSet
from ..fields import SparseField
from ..tags import USER, ITEM, ID
from ...dict2obj import Config


__all__ = ['AmazonBooks_m1', 'AmazonCDs_m1', 'AmazonMovies_m1', 'AmazonBeauty_m1', 'AmazonElectronics_m1']


class AmazonBooks_m1(ImplicitRecSet):
    """Amazon-Books dataset.
    |    Dataset     | #Users | #Items | #Interactions |  #Train   |  #Test  | Density |
    |:--------------:|:------:|:------:|:-------------:|:---------:|:-------:|:-------:|
    | AmazonBooks_m1 | 52,643 | 91,599 |   2,984,108   | 2,380,730 | 603,378 | 0.00062 |

    See [here](https://github.com/openbenchmark/BARS/tree/master/candidate_matching/datasets/Amazon#AmazonBooks_m1) for details.

    Attributes:
    ---

    _cfg: Config
        - sparse: SparseField
            UserID + ItemID
        - target: SparseField
            Rating
    open_kw: Config
        - mode: 'rt'
        - delimiter: ' '
        - skip_lines: 0

    """

    URL = "https://zenodo.org/record/7297855/files/AmazonBooks_m1.zip"

    _cfg = Config(
        sparse = [
            SparseField(name='UserID', na_value=None, dtype=int, tags=[USER, ID]),
            SparseField(name='ItemID', na_value=None, dtype=int, tags=[ITEM, ID]),
        ]
    )

    _cfg.fields = _cfg.sparse


class AmazonCDs_m1(ImplicitRecSet):
    """Amazon-CDs dataset.
    |   Dataset    | #Users | #Items | #Interactions | #Train  |  #Test  | Density |
    |:------------:|:------:|:------:|:-------------:|:-------:|:-------:|:-------:|
    | AmazonCDs_m1 | 43,169 | 35,648 |    777,426    | 604,475 | 172,951 | 0.00051 |

    See [here](https://github.com/openbenchmark/BARS/tree/master/candidate_matching/datasets) for details.

    Attributes:
    ---

    _cfg: Config
        - sparse: SparseField
            UserID + ItemID
        - target: SparseField
            Rating
    open_kw: Config
        - mode: 'rt'
        - delimiter: ' '
        - skip_lines: 0

    """

    URL = "https://zenodo.org/record/7297855/files/AmazonCDs_m1.zip"

    _cfg = Config(
        sparse = [
            SparseField(name='UserID', na_value=None, dtype=int, tags=[USER, ID]),
            SparseField(name='ItemID', na_value=None, dtype=int, tags=[ITEM, ID]),
        ]
    )

    _cfg.fields = _cfg.sparse


class AmazonMovies_m1(ImplicitRecSet):
    """Amazon-Movies dataset.
    |     Dataset     | #Users | #Items | #Interactions | #Train  |  #Test  | Density |
    |:---------------:|:------:|:------:|:-------------:|:-------:|:-------:|:-------:|
    | AmazonMovies_m1 | 44,439 | 25,047 |   1,070,860   | 839,444 | 231,416 | 0.00096 |

    See [here](https://github.com/openbenchmark/BARS/tree/master/candidate_matching/datasets) for details.

    Attributes:
    ---

    _cfg: Config
        - sparse: SparseField
            UserID + ItemID
        - target: SparseField
            Rating
    open_kw: Config
        - mode: 'rt'
        - delimiter: ' '
        - skip_lines: 0

    """

    URL = "https://zenodo.org/record/7297855/files/AmazonMovies_m1.zip"

    _cfg = Config(
        sparse = [
            SparseField(name='UserID', na_value=None, dtype=int, tags=[USER, ID]),
            SparseField(name='ItemID', na_value=None, dtype=int, tags=[ITEM, ID]),
        ]
    )

    _cfg.fields = _cfg.sparse


class AmazonBeauty_m1(ImplicitRecSet):
    """Amazon-Beauty dataset.
    |     Dataset     | #Users | #Items | #Interactions | #Train | #Test  | Density |
    |:---------------:|:------:|:------:|:-------------:|:------:|:------:|:-------:|
    | AmazonBeauty_m1 | 7,068  | 3,570  |    79,506     | 60,818 | 18,688 | 0.00315 |

    See [here](https://github.com/openbenchmark/BARS/tree/master/candidate_matching/datasets) for details.

    Attributes:
    ---

    _cfg: Config
        - sparse: SparseField
            UserID + ItemID
        - target: SparseField
            Rating
    open_kw: Config
        - mode: 'rt'
        - delimiter: ' '
        - skip_lines: 0

    """

    URL = "https://zenodo.org/record/7297855/files/AmazonBeauty_m1.zip"

    _cfg = Config(
        sparse = [
            SparseField(name='UserID', na_value=None, dtype=int, tags=[USER, ID]),
            SparseField(name='ItemID', na_value=None, dtype=int, tags=[ITEM, ID]),
        ]
    )

    _cfg.fields = _cfg.sparse


class AmazonElectronics_m1(ImplicitRecSet):
    """Amazon-Beauty dataset.
    |       Dataset        | #Users | #Items | #Interactions | #Train | #Test | Density |
    | :------------------: | :----: | :----: | :-----------: | :----: | :---: | :-----: |
    | AmazonElectronics_m1 | 1,435  | 1,522  |    35,931     | 31,887 | 4,044 | 0.01645 |
    See [here](https://github.com/xue-pai/UltraGCN/tree/main/data) for details.

    Attributes:
    ---

    _cfg: Config
        - sparse: SparseField
            UserID + ItemID
        - target: SparseField
            Rating
    open_kw: Config
        - mode: 'rt'
        - delimiter: ' '
        - skip_lines: 0

    """

    URL = "https://zenodo.org/record/7297855/files/AmazonElectronics_m1.zip"

    _cfg = Config(
        sparse = [
            SparseField(name='UserID', na_value=None, dtype=int, tags=[USER, ID]),
            SparseField(name='ItemID', na_value=None, dtype=int, tags=[ITEM, ID]),
        ]
    )

    _cfg.fields = _cfg.sparse




from enum import Enum


class FieldTags(Enum):
    Sparse = 'Sparse'
    Dense = 'Dense'
    ID = 'ID'
    User = 'User'
    Item = 'Item'
    Feature = 'Feature'
    Target = 'Target'


SPARSE = FieldTags('Sparse')
DENSE = FieldTags('Dense')
ID = FieldTags('ID')
USER = FieldTags('User')
ITEM = FieldTags('Item')
FEATURE = FieldTags('Feature')
TARGET = FieldTags('Target')




from typing import Dict, Optional, Union

import torch
import torch_geometric.transforms as T
from torch_geometric.data.data import Data
from torch_geometric.nn import LGConv

from .base import RecSysArch

from ..data.fields import Tokenizer
from ..data.tags import ID, ITEM, USER


__all__ = ['LightGCN']


class LightGCN(RecSysArch):

    def __init__(
        self, tokenizer: Tokenizer, 
        graph: Data,
        num_layers: int = 3
    ) -> None:
        super().__init__()

        self.tokenizer = tokenizer
        self.conv = LGConv()
        self.num_layers = num_layers
        self.graph = graph

        self.User, self.Item = self.tokenizer[USER, ID], self.tokenizer[ITEM, ID]

        self.initialize()

    @property
    def graph(self):
        return self.__graph

    @graph.setter
    def graph(self, graph: Data):
        self.__graph = graph
        T.ToSparseTensor()(self.__graph)

    def to(
        self, device: Optional[Union[int, torch.device]] = None, 
        dtype: Optional[Union[torch.dtype, str]] = None, 
        non_blocking: bool = False
    ):
        if device:
            self.graph.to(device)
        return super().to(device, dtype, non_blocking)

    def forward(
        self, users: Optional[Dict[str, torch.Tensor]] = None, 
        items: Optional[Dict[str, torch.Tensor]] = None
    ):
        userEmbs = self.User.embeddings.weight
        itemEmbs = self.Item.embeddings.weight
        features = torch.cat((userEmbs, itemEmbs), dim=0).flatten(1) # N x D
        avgFeats = features / (self.num_layers + 1)
        for _ in range(self.num_layers):
            features = self.conv(features, self.graph.adj_t)
            avgFeats += features / (self.num_layers + 1)
        userFeats, itemFeats = torch.split(avgFeats, (self.User.count, self.Item.count))

        if self.training: # Batch
            users, items = self.broadcast(
                users[self.User.name], items[self.Item.name]
            )
            userFeats = userFeats[users] # B x n x D
            itemFeats = itemFeats[items] # B x n x D
            userEmbs = self.User.look_up(users) # B x n x D
            itemEmbs = self.Item.look_up(items) # B x n x D
            return torch.mul(userFeats, itemFeats).sum(-1), userEmbs, itemEmbs
        else:
            return userFeats, itemFeats





from typing import Callable, Iterator, Optional, Dict, Union

import torch
import torchdata.datapipes as dp
import numpy as np
from math import ceil

from freerec.utils import errorLogger

from .base import Postprocessor


__all__ = [
    "Sharder", "Shuffle", "SubFielder", "Spliter", "ToTensor"
]


@dp.functional_datapipe("shard_")
class Sharder(Postprocessor):
    """For num_workers != 0."""

    def __iter__(self) -> Iterator:
        worker_infos = torch.utils.data.get_worker_info()
        if worker_infos:
            id_, nums = worker_infos.id, worker_infos.num_workers
            for idx, item in enumerate(self.source):
                if idx % nums == id_:
                    yield item
        else:
            yield from self.source


@dp.functional_datapipe("shuffle_")
class Shuffle(Postprocessor):
    """Shuffling the datapipe."""

    def shuffle(self, chunk: Dict):
        indices = None
        for key in chunk:
            if indices is None:
                indices = np.arange(len(chunk[key]))
                np.random.shuffle(indices)
            chunk[key] = chunk[key][indices]
        return chunk

    def __iter__(self):
        for chunk in self.source:
            yield self.shuffle(chunk)


@dp.functional_datapipe("subfield_")
class SubFielder(Postprocessor):
    """Select subfields."""

    def __init__(
        self, datapipe: Postprocessor,
        filter_: Optional[Callable] = None
    ) -> None:
        super().__init__(datapipe)
        if filter_:
            self.fields = [field for field in self.fields if filter_(field)]

    def __iter__(self) -> Iterator:
        for chunk in self.source:
            yield {field.name: chunk[field.name] for field in self.fields}


@dp.functional_datapipe("tensor_")
class ToTensor(Postprocessor):
    """Convert Dict[str, List] into Dict[str, torch.Tensor]."""

    def at_least_2d(self, vals: torch.Tensor):
        return vals.unsqueeze(1) if vals.ndim == 1 else vals

    def __iter__(self) -> Iterator:
        for batch in self.source:
            yield {field.name: self.at_least_2d(torch.tensor(batch[field.name], dtype=field.dtype)) for field in self.fields}


@dp.functional_datapipe("split_")
class Spliter(Postprocessor):
    """A special batcher for Dict[str, Union[Tensor, ndarray]] only."""

    def __init__(
        self, datapipe: Postprocessor, batch_size: int
    ) -> None:
        """
        Parameters:
        ---

        datapipe: Postprocessor
            It must yield tensors or array !
        batch_size: int
            If the given batch size is not evenly divisible by _DEFAULT_CHUNK_SIZE,
            minor chunk will yielded frequently. Fortunately, _DEFAULT_CHUNK_SIZE=51200 satisfies
            almost all popular batch sizes. For example, 128, 1024 ... 
        """
        super().__init__(datapipe)

        self.batch_size = batch_size

    def split(self, vals: Union[torch.Tensor, np.ndarray]):
        if isinstance(vals, torch.Tensor):
            return torch.split(vals, self.batch_size, dim=0)
        elif isinstance(vals, np.ndarray):
            return np.array_split(vals, ceil(len(vals) / self.batch_size), axis=0)
        else:
            errorLogger(f"Only Tensor or ndarray supported but {type(vals)} received", ValueError)

    def __len__(self):
        raise errorLogger("Spliter has no `__len__` method ...", NotImplementedError)

    def __iter__(self) -> Iterator:
        for chunk in self.source:
            chunk = {key: self.split(vals) for key, vals in chunk.items()}
            try:
                k = 0
                while True:
                    yield {key: chunk[key][k] for key in chunk}
                    k += 1
            except IndexError:
                pass






from typing import Iterator, Iterable, Union, List, Optional

import torchdata.datapipes as dp

from .base import Postprocessor
from ..tags import FieldTags, USER, ITEM, TARGET


__all__ = ['Grouper', 'Wrapper']


@dp.functional_datapipe("group_")
class Grouper(Postprocessor):
    """Group batch into several groups."""
    def __init__(
        self, datapipe: Postprocessor, 
        groups: Iterable[Union[FieldTags, Iterable[FieldTags]]] = (USER, ITEM, TARGET)
    ) -> None:
        """
        Parameters:
        ---
        groups: Iterable[Union[FieldTags, Iterable[FieldTags]]]
            Gathering fields for each group of tags in groups.

        Examples:
        ---

        >>> from freerec.data.tags import SPARSE, TARGET, USER, ITEM
        >>> from freerec.data.datasets import MovieLens1M
        >>> basepipe = MovieLens1M("../../data/MovieLens1M")
        >>> datapipe = basepipe.tensor_().chunk_(1024)
        >>> dataset = datapipe.wrap_().group_((USER, ITEM, TARGET))
        >>> len(next(iter(dataset)))
        3
        >>> dataset = datapipe.wrap_().group_((ID, TARGET))
        >>> len(next(iter(dataset)))
        2
        """
        super().__init__(datapipe)
        self.groups = []
        for tags in groups:
            if isinstance(tags, FieldTags):
                tags = (tags,)
            self.groups.append([field for field in self.fields if field.match(*tags)])

    def __iter__(self) -> List:
        for batch in self.source:
            yield [{field.name: batch[field.name] for field in group} for group in self.groups]


@dp.functional_datapipe("wrap_")
class Wrapper(Postprocessor):
    """Wrap trainpipe, validpipe and testpipe together."""

    def __init__(
        self, datapipe: Postprocessor,
        validpipe: Optional[Postprocessor] = None,
        testpipe: Optional[Postprocessor] = None,
    ) -> None:
        """
        Parameters:
        ---

        datapipe: trainpipe
        validpipe: validpipe or None
            - `None`: Set validpipe == trainpipe

        testpipe: testpipe
            - `None`: Set testpipe == validpipe

        Examples:
        ---

        >>> from freerec.data.datasets import MovieLens1M
        >>> basepipe = MovieLens1M("../../data/MovieLens1M")
        >>> datapipe = basepipe.shuffle_().shard_()
        >>> trainpipe = datapipe.negatives_for_train_(num_negatives=4)
        >>> validpipe = datapipe.negatives_for_eval_(num_negatives=99) # 1:99
        >>> dataset = trainpipe.wrap_(validpipe).tensor_().chunk_(cfg.batch_size).group_()
        """
        super().__init__(datapipe)
        self.validpipe = datapipe if validpipe is None else validpipe
        self.testpipe = self.validpipe if testpipe is None else testpipe
        self.train()

    def train(self):
        super().train()
        self.source.train()
        return self

    def valid(self):
        super().valid()
        self.validpipe.valid()
        return self

    def test(self):
        super().test()
        self.testpipe.test()
        return self

    def __len__(self):
        if self.mode == 'train':
            return len(self.source)
        elif self.mode == 'valid':
            return len(self.validpipe)
        else:
            return len(self.testpipe)

    def __iter__(self) -> Iterator:
        if self.mode == 'train':
            yield from self.source
        elif self.mode == 'valid':
            yield from self.validpipe
        else:
            yield from self.testpipe





from typing import Any, Callable, Iterable, List, Dict, Optional, Tuple, Union

import torch
import pandas as pd
import os, subprocess, shlex, time, sys
from torch.utils.tensorboard import SummaryWriter
from functools import partial
from itertools import product
from collections import defaultdict
from tqdm import tqdm
from freeplot.utils import import_pickle, export_pickle

from .data.datasets.base import BaseSet
from .data.fields import Field, Fielder
from .data.dataloader import DataLoader
from .models import RecSysArch
from .criterions import BaseCriterion
from .dict2obj import Config
from .utils import AverageMeter, Monitor, timemeter, infoLogger, errorLogger
from .metrics import *
from .parser import TIME


__all__ = ['Coach', 'Adapter']


DEFAULT_METRICS = {
    'LOSS': None,
    #############
    'MSE': mean_squared_error,
    'MAE': mean_abs_error,
    'RMSE': root_mse,
    #############
    'PRECISION': precision,
    'RECALL': recall,
    'F1': f1_score,
    'AUC': auroc,
    'HITRATE': hit_rate,
    #############
    'NDCG': normalized_dcg,
    'MRR': mean_reciprocal_rank,
    'MAP': mean_average_precision
}

DEFAULT_FMTS = {
    'LOSS': ".5f",
    #############
    'MSE': ".4f",
    'MAE': ".4f",
    'RMSE': ".4f",
    #############
    'PRECISION': ".4f",
    'RECALL': ".4f",
    'F1': ".4f",
    'AUC': ".4f",
    'HITRATE': ".4f",
    #############
    'NDCG': ".4f",
    'MRR': ".4f",
    'MAP': ".4f",
}

DEFAULT_BEST_CASTER = {
    'LOSS': min,
    #############
    'MSE': min,
    'MAE': min,
    'RMSE': min,
    #############
    'PRECISION': max,
    'RECALL': max,
    'F1': max,
    'AUC': max,
    'HITRATE': max,
    #############
    'NDCG': max,
    'MRR': max,
    'MAP': max,
}

class _DummyModule(torch.nn.Module):

    def forward(self, *args, **kwargs):
        errorLogger("No model available for Coach ...", NotImplementedError)

    def step(self, *args, **kwargs):
        errorLogger("No optimizer or lr scheduler available for Coach ...", NotImplementedError)

    def backward(self, *args, **kwargs):
        errorLogger("No optimizer available for Coach ...", NotImplementedError)


class Coach:
    """The framework for training."""
    
    def __init__(
        self, 
        dataset: BaseSet,
        model: Union[RecSysArch, torch.nn.Module, None],
        criterion: Union[BaseCriterion, Callable],
        optimizer: Optional[torch.optim.Optimizer],
        lr_scheduler: Optional[torch.optim.lr_scheduler._LRScheduler],
        device: Union[torch.device, str, int]
    ):
        """
        Parameters:
        ---

        dataset: RecDataSet for training, validation and test
        criterion: Callable
            Loss funcion.
        model: nn.Module or None
            - `None`: Using _DummyModule instead, by which `forward` should not be called.
        
        optimizer: torch.optim.Optimizer or None
            - `None`: Using _DummyModule instead, by which `step` and `backward` should not be called.

        lr_scheduler: Callable or None
            - `None`: Using _DummyModule instead, by which `step` should not be called.

        device: torch.device, str or int
            - `torch.device`
            - `str`: Like `cpu`, `cuda:0`.
            - `int`: Using cuda:`int`.
        """
        self.dataset = dataset
        self.fields: Fielder[Field] = self.dataset.fields
        self.device = torch.device(device)
        self.criterion = criterion
        self.model = model.to(self.device) if model else _DummyModule()
        self.optimizer = optimizer if optimizer else _DummyModule()
        self.lr_scheduler = lr_scheduler if lr_scheduler else _DummyModule()

      
    def save(self) -> None:
        """Save the model"""
        torch.save(self.model.state_dict(), os.path.join(self.cfg.LOG_PATH, self.cfg.SAVED_FILENAME))

    def save_checkpoint(self, epoch: int) -> None:
        """Save current checkpoint at epoch X."""
        path = os.path.join(self.cfg.CHECKPOINT_PATH, self.cfg.CHECKPOINT_FILENAME)
        checkpoint = dict()
        checkpoint['epoch'] = epoch
        for module in self.cfg.CHECKPOINT_MODULES:
            checkpoint[module] = getattr(self, module).state_dict()
        checkpoint['monitors'] = self.monitors.state_dict()
        torch.save(checkpoint, path)

    def load_checkpoint(self) -> int:
        """Load last saved checkpoint.

        Returns:
        ---

        epoch: int
        """
        path = os.path.join(self.cfg.CHECKPOINT_PATH, self.cfg.CHECKPOINT_FILENAME)
        checkpoint = torch.load(path)
        for module in self.cfg.CHECKPOINT_MODULES:
            getattr(self, module).load_state_dict(checkpoint[module])
        self.monitors.load_state_dict(checkpoint['monitors'])
        return checkpoint['epoch']

    def save_best(self, path: str, prefix: str): ...

    def check_best(self, val: Optional[float]): ...

    @timemeter("Coach/resume")
    def resume(self):
        """Resume from last checkpoint."""
        start_epoch = self.load_checkpoint() if self.cfg.resume else 0
        infoLogger(f"[Coach] >>> Load the recent checkpoint and train from epoch: {start_epoch}")
        return start_epoch
    
    def register_metric(
        self, name: str, metric: Optional[str] = None, 
        func: Optional[Callable] = None, fmt: str = '.4f', best_caster: Callable = max,
        prefix: str = 'train'
    ):
        """Register a metric.

        Parameters:
        ---

        name: str
            The complete name of this metric, such as `NDCG@20`.
        metric: str
            The name of this metric, like `NDCG` for `NDCG@20`.
        func: Callable
            The function to process data.
        fmt: str
            The format to print.
        best_caster: `min` or `max`
            Which one is better.
        prefix: str
            The group to which this metric belongs.

        Examples
        ---

        >>> coach = Coach(...)
        >>> coach.compile(cfg, monitors=['loss', 'recall@10', 'recall@20', 'ndcg@10', 'ndcg@20'])
        >>> coach.register_metric('loss2', func=None, fmt='.5f', prefix='train')
        >>> from freerec.metrics import normalized_dcg
        >>> from functools import partial
        >>> coach.register_metric(
        ...    name='NDCG@50',
        ...    metric='NDCG',
        ...    func=partial(normalized_dcg, k=50),
        ...   fmt='.4f',
        ...    prefix='test'
        ... )

        Raises:
        ---

        AttributeError: when calling `register_metric' before `compile'.
        
        """
        try:
            metric = name if metric is None else metric
            self.__monitors[prefix][metric.upper()].append(
                AverageMeter(
                    name=name,
                    metric=func,
                    fmt=fmt,
                    best_caster=best_caster
                )
            )
        except AttributeError:
            errorLogger(
                "'register_metric' should be called after 'compile' ...",
                AttributeError
            )

    @property
    def monitors(self) -> Monitor:
        """Return Dict[str, Dict[str, List]], specifically,
        Monitor[prefix: str, Dict[metric: str, meters: List]].
        """
        return self.__monitors

    @timemeter("Coach/compile")
    def compile(
        self, cfg: Config, monitors: List[str]
    ):
        """Load config and set monitors.
        
        Parameters:
        ---

        cfg: Config
        monitors: List[str]
            The metrics (for 'valid' and 'test' only) of interest.
        
        Examples:
        ---

        >>> coach = Coach(None)
        >>> coach.compile(cfg, monitors=['loss', 'recall@10', 'recall@20', 'ndcg@10', 'ndcg@20'])
        """
        self.cfg = cfg
        # meters for train|valid|test
        self.__monitors = Monitor()
        self.__monitors['train'] = defaultdict(list)
        self.__monitors['valid'] = defaultdict(list)
        self.__monitors['test'] = defaultdict(list)

        self.register_metric(
            name='LOSS',
            metric='LOSS',
            func=DEFAULT_METRICS['LOSS'],
            fmt=DEFAULT_FMTS['LOSS'],
            best_caster=DEFAULT_BEST_CASTER['LOSS'],
            prefix='train'
        )

        for name in monitors:
            name = name.upper()
            if '@' in name:
                metric, K = name.split('@')
                for prefix in ('valid', 'test'):
                    self.register_metric(
                        name=name,
                        metric=metric,
                        func=partial(DEFAULT_METRICS[metric], k=int(K)),
                        fmt=DEFAULT_FMTS[metric],
                        best_caster=DEFAULT_BEST_CASTER[metric],
                        prefix=prefix
                    )
            else:
                metric = name
                for prefix in ('valid', 'test'):
                    self.register_metric(
                        name=name,
                        metric=metric,
                        func=DEFAULT_METRICS[metric],
                        fmt=DEFAULT_FMTS[metric],
                        best_caster=DEFAULT_BEST_CASTER[metric],
                        prefix=prefix
                    )

        # dataloader
        self.load_dataloader()

    def load_dataloader(self):
        self.__dataloader = DataLoader(
            datapipe=self.dataset, num_workers=self.cfg.num_workers
        )
    
    @property
    def dataloader(self):
        if self.cfg.verbose:
            return tqdm(
                self.__dataloader,
                leave=False, desc="վ'ᴗ' ի-"
            )
        else:
            return self.__dataloader

    @torch.no_grad()
    def monitor(
        self, *values,
        n: int = 1, mode: str = 'mean', 
        prefix: str = 'train', pool: Optional[List] = None
    ):
        """Log some values to given monitors.

        Parameters:
        ---

        *values: data
        n: int
            Batch size in general
        mode: 'sum'|'mean' (default)
        prefix: str, 'train'|'test'|'valid'
            The mode values belonging to.
        pool: List[str]
            Given metrics.
            - `None`: `pool` will be set for all metrics.
        """
        metrics: Dict[List] = self.monitors[prefix]
        pool = metrics if pool is None else pool
        for metric in pool:
            for meter in metrics.get(metric.upper(), []):
                meter(*values, n=n, mode=mode)

    def step(self, epoch: int):
        """Print information and reset them."""
        for prefix, metrics in self.monitors.items():
            metrics: Dict[str, List[AverageMeter]]
            infos = [f"[Coach] >>> {prefix.upper():5} @Epoch: {epoch:<4d} >>> "]
            for meters in metrics.values():
                infos += [meter.step() for meter in meters if meter.active]
            infoLogger(' || '.join(infos))

    @timemeter("Coach/summary")
    def summary(self):
        """Summary the whole training process.
        The following information will be saved:
            1. historical evaluation saved in monitors;
            2. best historical results;
            3. curves of historical results.
        """
        file_ = os.path.join(self.cfg.LOG_PATH, self.cfg.SUMMARY_DIR, self.cfg.SUMMARY_FILENAME)
        s = "|  {prefix}  |   {metric}   |   {val}   |   {epoch}   |   {img}   |\n"
        info = ""
        info += "|  Prefix  |   Metric   |   Best   |   @Epoch   |   Img   |\n"
        info += "| :-------: | :-------: | :-------: | :-------: | :-------: |\n"
        data = []
        best = defaultdict(dict)

        for prefix, metrics in self.monitors.items():
            metrics: defaultdict[str, List[AverageMeter]]
            freq = 1 if prefix == 'train' else self.cfg.EVAL_FREQ
            for _, meters in metrics.items():
                for meter in meters:
                    meter.plot(freq=freq)
                    imgname = meter.save(path=os.path.join(self.cfg.LOG_PATH, self.cfg.SUMMARY_DIR), prefix=prefix)
                    epoch, val = meter.argbest(freq)
                    info += s.format(
                        prefix=prefix, metric=meter.name,
                        val=val, epoch=epoch, img=f"![]({imgname})"
                    )
                    data.append([prefix, meter.name, val, epoch])
                    if val != -1: # Only save available data.
                        best[prefix][meter.name] = val

        with open(file_, "w", encoding="utf8") as fh:
            fh.write(info) # Summary.md

        df = pd.DataFrame(data, columns=['Prefix', 'Metric', 'Best', '@Epoch'])
        infoLogger(str(df)) # print final metrics
        infoLogger(f"[LoG_PaTH] >>> {self.cfg.LOG_PATH}")
        # save corresponding data for next analysis
        self.monitors.write(os.path.join(self.cfg.LOG_PATH, self.cfg.SUMMARY_DIR)) # tensorboard
        self.monitors.save(os.path.join(self.cfg.LOG_PATH, self.cfg.DATA_DIR), self.cfg.MONITOR_FILENAME)
        export_pickle(best, os.path.join(self.cfg.LOG_PATH, self.cfg.DATA_DIR, self.cfg.MONITOR_BEST_FILENAME))

    def train_per_epoch(self):
        errorLogger("train_per_epoch should be specified ...", NotImplementedError)

    def evaluate(self, prefix: str = 'valid'):
        errorLogger("evaluate should be specified ...", NotImplementedError)

    @timemeter("Coach/train")
    def train(self):
        self.dataset.train()
        self.model.train()
        return self.train_per_epoch()

    @timemeter("Coach/valid")
    @torch.no_grad()
    def valid(self):
        self.dataset.valid() # TODO: multiprocessing pitfall ???
        self.model.eval()
        return self.evaluate(prefix='valid')

    @timemeter("Coach/test")
    @torch.no_grad()
    def test(self):
        self.dataset.test()
        self.model.eval()
        return self.evaluate(prefix='test')

    @timemeter("Coach/fit")
    def fit(self):
        """Start the training and log some useful information."""
        start_epoch = self.resume()
        for epoch in range(start_epoch, self.cfg.epochs):
            if epoch % self.cfg.CHECKPOINT_FREQ == 0:
                self.save_checkpoint(epoch)
            if epoch % self.cfg.EVAL_FREQ == 0:
                if self.cfg.EVAL_VALID:
                    self.check_best(self.valid())
                if self.cfg.EVAL_TEST:
                    self.test()
            self.train()

            self.step(epoch)
        self.save()

        # last epoch
        self.check_best(self.valid())
        self.test()
        self.step(self.cfg.epochs)

        self.summary()


class CoachForCTR(Coach): ...

class CoachForMatching(Coach): ...


class Adapter:
    """ Params tuner.

    Flows:
    ---

    1. compile: get command, envs and params for training;
    2. allocate devices for various params
        - register id, logPath, device first
        - run it
        - collect information from logPath and output to tensorbaord
        - save checkpoint
        - release corresponding device

    Examples:
    ---

    >>> cfg = {'command': 'python xxx.py', 'params': {'optimizer': ['sgd', 'adam']}}
    >>> tuner = Adapter()
    >>> tuner.compile(cfg)
    >>> tuner.fit()
    """

    def __init__(self) -> None:
        self.params = []
        self.values = []
        self.devices = []

    @property
    def COMMAND(self):
        return self.cfg.COMMAND

    def register(self, device: str) -> Tuple[str, str]:
        self.cfg.ENVS['id'] = time.strftime(TIME)
        self.cfg.ENVS['device'] = device
        command = self.COMMAND + self.get_option('id', self.cfg.ENVS.id)
        command += self.get_option('device', self.cfg.ENVS.device)
        return command, self.cfg.ENVS.id, self.cfg.LOG_PATH.format(**self.cfg.ENVS)

    @timemeter("Adapter/compile")
    def compile(self, cfg: Config):
        """
        Parameters:
        ---

        cfg: Config
            Including command, envs, params, defaults.
        
        Flows:
        ---

        1. Add environmental parameters to basic `command`;
        2. Register all available devices;
        3. Convert all parameters from `cfg.PARAMS`;
        4. Convert all defaults from `cfg.DEFAULTS`.
        """
        self.cfg = cfg
        piece = "\t{key}: {vals} \n"
        envs, params, defaults = "", "", ""
        for key, val in self.cfg.ENVS.items():
            if key == 'device':
                self.devices = val.split(',')
            else:
                self.cfg.COMMAND += self.get_option(key, val)
            envs += piece.format(key=key, vals=val)
        for key, vals in self.cfg.PARAMS.items():
            if isinstance(vals, (str, int, float)):
                vals = [vals]
            self.deploy_params(key, vals)
            params += piece.format(key=key, vals=vals)
        for key, val in self.cfg.DEFAULTS.items():
            self.cfg.DEFAULTS[key] = val
            defaults += piece.format(key=key, vals=val)

        cfg_infos = f"command: {self.cfg.COMMAND} \nenvs: \n{envs}params: \n{params}defaults: \n{defaults}"
        infoLogger(f"\033[0;31;47m{cfg_infos}\033[0m")
        

    def deploy_params(self, key: str, vals: Iterable):
        self.params.append(key)
        self.values.append(vals)

    @staticmethod
    def get_option(key: str, val: Any):
        """Convert (key, val) to '--key=val'.

        Parameters:
        ---

        key: str
        val: Any

        Notes:
        ---

        All '_' in key will be replaced by '-'.
        
        Examples:
        ---

        >>> Adapter.get_option('lr', '1e-3')
        --lr=1e-3
        >>> Adapter.get_option('learning_rate', '1e-3')
        --learning-rate=1e-3
        """
        return f" --{key.replace('_', '-')}={val}"

    def load_best(self, logPath: str):
        """Load best.pickle from logPath of corresponding."""
        file_ = os.path.join(logPath, self.cfg.DATA_DIR, self.cfg.MONITOR_BEST_FILENAME)
        return import_pickle(file_)

    def write(self, id_: str, logPath: str, params: Dict):
        """Write results to tensorboard.
        
        Parameters:
        ---

        id_: str
            The experiment id.
        logPath: str
            The log path of this experiment.
        params: Dict
            Parameter config of this experiemnt.
        
        Flows:
        ---

        1. Load best data from the `logPath`.
        2. Write best data to tensorboard with `params`.

        Notes:
        ---

        If you find `-1` appears in the tensorboard,
        it must be the data therein is of `str` type,
        which will raise error if we sent it to tensorboard directly !
        """
        data = self.load_best(logPath)
        path = os.path.join(self.cfg.CORE_LOG_PATH, id_)
        with SummaryWriter(log_dir=path) as writer:
            metrics = dict()
            for prefix, best in data.items():
                for metric, val in best.items():
                    val = val if isinstance(val, (int, float)) else -1
                    metrics['/'.join([prefix, metric])] = val
            writer.add_hparams(
                params, metrics,
            )

    def each_grid(self):
        """Grid search for each kind of param."""
        for key, vals in zip(self.params, self.values):
            for val in vals:
                yield self.cfg.DEFAULTS | {key: val}

    def product_grid(self):
        """Grid search across all combination of params"""
        for vals in product(*self.values):
            yield self.cfg.DEFAULTS | {option:val for option, val in zip(self.params, vals)}

    def save_checkpoint(self, source: List) -> None:
        """Save the rest of params."""
        path = os.path.join(self.cfg.CORE_CHECKPOINT_PATH, self.cfg.CHECKPOINT_FILENAME)
        checkpoint = dict()
        checkpoint['source'] = source
        torch.save(checkpoint, path)

    def load_checkpoint(self) -> int:
        """Load the rest of params."""
        path = os.path.join(self.cfg.CORE_CHECKPOINT_PATH, self.cfg.CHECKPOINT_FILENAME)
        checkpoint = torch.load(path)
        return checkpoint['source']

    @timemeter("Coach/resume")
    def resume(self):
        """Resume from the recent checkpoint."""
        source = self.each_grid() if self.cfg.EXCLUSIVE else self.product_grid()
        source = list(source)[::-1]
        source = self.load_checkpoint() if self.cfg.resume else source
        infoLogger(f"[Coach] >>> Load the recent checkpoint ...")
        return source

    def run(self, command: str, params: Dict):
        """Start a new subprocess"""
        for option, val in params.items():
            command += self.get_option(option, val)
        infoLogger(f"\033[0;31;47m{command}\033[0m")
        return subprocess.Popen(shlex.split(command))

    def wait(self, tasks: Dict):
        """Wait util all processes terminate."""
        for process_, id_, logPath, params in tasks.values():
            process_.wait()
            self.write(id_, logPath, params)

    def poll(self, tasks: Dict):
        """Wait util any process terminates."""
        buffer_source = []
        time.sleep(1) # for unique id
        while len(self.devices) == 0:
            time.sleep(7)
            for device, (process_, id_, logPath, params) in tasks.items():
                if process_.poll() is not None:
                    self.write(id_, logPath, params)
                    self.devices.append(device)
                else:
                    buffer_source.append(params)
        self.save_checkpoint(self.source + buffer_source)

    @timemeter("Adapter/fit")
    def fit(self):
        """Grid search."""

        self.source = self.resume()
        tasks = dict()
        try:
            while self.source:
                self.poll(tasks)
                params = self.source.pop()
                device = self.devices.pop()
                command, id_, logPath = self.register(device)
                process_ = self.run(command, params)
                tasks[device] = (process_, id_, logPath, params)
        except Exception as e:
            print(e)
        finally:
            self.wait(tasks)
            sys.exit()




from typing import Callable, Optional, Dict, List, Union
import torch
import numpy as np
from torch.utils.tensorboard import SummaryWriter

import logging, time, random, os
from collections import defaultdict
from freeplot.base import FreePlot
from freeplot.utils import export_pickle

from .dict2obj import Config


class FreeRecError(Exception): ...

LOGGER = Config(
    name='RecSys', filename='log.txt', level=logging.DEBUG,
    filelevel=logging.DEBUG, consolelevel=logging.INFO,
    formatter=Config(
        filehandler=logging.Formatter('%(asctime)s:\t%(message)s'),
        consolehandler=logging.Formatter('%(message)s')
    ),
    info=print,
    debug=print
)

COLOR = {
    'current': "{0}",
    'cpu': "{0}",
    0: "{0}",
    1: "\033[1;35m{0}\033[0m",
    2: "\033[1;34m{0}\033[0m",
    3: "\033[1;33m{0}\033[0m",
    4: "\033[1;38m{0}\033[0m",
    5: "\033[1;32m{0}\033[0m"
}


def _unitary(value):
    return value

class AverageMeter:

    def __init__(
        self, name: str, metric: Optional[Callable] = None, 
        fmt: str = ".5f", best_caster: Callable = max
    ):
        """
        Paramters:
        ---

        name: The name of the meter.
        metric: Metrics from freerec.metrics.
        fmt: Output format (default: ".5f").
        best_caster: min or max 
            Depends on which indicates better performance for this metric.
        """
        self.name = name
        self.fmt = fmt
        self.caster = best_caster
        self.reset()
        self.__history = []
        self.__metric = metric if metric else _unitary

    @property
    def history(self):
        """Return historical results."""
        return self.__history

    @history.setter
    def history(self, val: List):
        self.__history = val.copy()

    def reset(self) -> None:
        self.val = 0.
        self.avg = 0.
        self.sum = 0.
        self.count = 0
        self.active = False

    def update(self, val: float, n: int = 1, mode: str = "mean") -> None:
        """Update accumulatively.

        Parameters:
        ---

        val: Value.
        n: Batch size in general.
        mode: 'sum'|'mean'(default).
        """
        self.val = val
        self.count += n
        if mode == "mean":
            self.sum += val * n
        elif mode == "sum":
            self.sum += val
        else:
            raise ValueError(f"Receive mode {mode} but 'mean' or 'sum' expected ...")
        self.avg = self.sum / self.count

    def step(self) -> str:
        """Move current value into the queue and reset the state."""
        self.history.append(self.avg)
        info = str(self)
        self.reset()
        return info

    def check(self, *values):
        val = self.__metric(*values)
        if isinstance(val, torch.Tensor):
            val = val.cpu().numpy()
        else:
            val = np.array(val)
        if np.isnan(val) or np.isinf(val):
            errorLogger(
                f"The metric of {self.name} got an unexpected value: {val.item()}."
            )
        return val.item()

    def plot(self, freq: int = 1) -> None:
        """Plot lines according to historical results."""
        self.fp = FreePlot(
            shape=(1, 1),
            titles=(self.name,),
            dpi=300, latex=False
        )
        timeline = np.arange(len(self.history)) * freq
        self.fp.lineplot(timeline, self.history, marker='')
        self.fp.set_title(y=.98)

    def save(self, path: str, prefix: str = '') -> None:
        """Save the curves."""
        filename = f"{prefix}{self.name}.png"
        self.fp.savefig(os.path.join(path, filename))
        return filename

    def argbest(self, freq: int = 1) -> float:
        """Return (whichisbest, best) in history.

        Paramters:
        ---

        freq: EVAL_FREQ for T * EVAL_FREQ.

        Returns:
        ---

        index: The index of best result.
            - `-1`: No available data.

        value: The best result.
            - `-1`: No available data.
        """
        if len(self.history) == 0:
            return -1, -1
        indices = np.argsort(self.history)
        if self.caster is min:
            return indices[0] * freq, self.history[indices[0]]
        elif self.caster is max:
            return indices[-1] * freq, self.history[indices[-1]]
        else:
            raise ValueError("caster should be min or max ...")

    def __str__(self):
        fmtstr = "{name} Avg: {avg:{fmt}}"
        return fmtstr.format(**self.__dict__)

    def __call__(self, *values, n: int = 1, mode: str = "mean")  -> None:
        self.active = True
        self.update(
            val = self.check(*values),
            n = n,
            mode = mode
        )


class Monitor(Config):

    def state_dict(self) -> Dict:
        """Return the state dict of monitors."""
        state_dict = defaultdict(dict)
        monitors: Dict[str, List[AverageMeter]]
        for prefix, monitors in self.items():
            for metric, meters in monitors.items():
                state_dict[prefix][metric] = dict()
                for meter in meters:
                    state_dict[prefix][metric][meter.name] = meter.history
        return state_dict

    def load_state_dict(self, state_dict: Dict, strict: bool = False):
        monitors: Dict[str, List[AverageMeter]]
        for prefix, monitors in self.items():
            for metric, meters in monitors.items():
                for meter in meters:
                    meter.history = state_dict[prefix][metric].get(meter.name, meter.history)

    def save(self, path: str, filename: str = 'monitors.pickle'):
        """Save current state."""
        file_ = os.path.join(path, filename)
        export_pickle(self.state_dict(), file_)

    def write(self, path: str):
        """Write to tensorboard."""
        with SummaryWriter(path) as writer:
            monitors: Dict[str, List[AverageMeter]]
            for prefix, monitors in self.items():
                for metric, meters in monitors.items():
                    for meter in meters:
                        for t, val in enumerate(meter.history):
                            writer.add_scalar(
                                '/'.join([prefix, meter.name]),
                                val,
                                t
                            )


def set_logger(
    path: str,
    log2file: bool = True, log2console: bool = True
) -> None:
    logger = logging.getLogger(LOGGER.name)
    logger.setLevel(LOGGER.level)

    if log2file:
        handler = logging.FileHandler(
            os.path.join(path, LOGGER.filename), 
            encoding='utf-8'
        )
        handler.setLevel(LOGGER.filelevel)
        handler.setFormatter(LOGGER.formatter.filehandler)
        logger.addHandler(handler)
    if log2console:
        handler = logging.StreamHandler()
        handler.setLevel(LOGGER.consolelevel)
        handler.setFormatter(LOGGER.formatter.consolehandler)
        logger.addHandler(handler)
    logger.debug("========================================================================")
    logger.debug("========================================================================")
    logger.debug("========================================================================")
    LOGGER['info'] = logger.info
    LOGGER['debug'] = logger.debug
    return logger

def set_color(device: Union[int, str]):
    """Set a color for current device."""
    try:
        COLOR['current'] = COLOR[device]
    except KeyError:
        pass

def infoLogger(words: str):
    words = COLOR['current'].format(words)
    LOGGER.info(words)
    return words

def debugLogger(words: str):
    LOGGER.debug(words)
    return words

def warnLogger(warn: str):
    words = f"\033[1;31m[Warning] >>> {warn} \033[0m"
    LOGGER.info(words)
    return words

def errorLogger(error: str, exception = FreeRecError):
    words = f"\033[1;31m{error} \033[0m"
    LOGGER.debug(words)
    raise exception(words)

def timemeter(prefix=""):
    def decorator(func):
        def wrapper(*args, **kwargs):
            start = time.time()
            results = func(*args, **kwargs)
            end = time.time()
            infoLogger(f"[Wall TIME] >>> {prefix} takes {end-start:.6f} seconds ...")
            return  results
        wrapper.__doc__ = func.__doc__
        wrapper.__name__ = func.__name__
        return wrapper
    return decorator

def mkdirs(*paths: str) -> None:
    for path in paths:
        try:
            os.makedirs(path)
        except FileExistsError:
            pass

def activate_benchmark(benchmark: bool) -> None:
    from torch.backends import cudnn
    if benchmark:
        infoLogger(f"[Benchmark] >>> cudnn.deterministic == True")
        cudnn.benchmark, cudnn.deterministic = False, True
    else:
        infoLogger(f"[Benchmark] >>> cudnn.deterministic == False")
        cudnn.benchmark, cudnn.deterministic = True, False

def set_seed(seed: int) -> int:
    """Set seed.

    Parameters:
    seed: int
        - `-1`: Set a seed from 0 to 2048 randomly.
        - `int`: Set a seed of `int`.

    Returns:
    seed: int
    """
    if seed == -1:
        seed = random.randint(0, 2048)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    infoLogger(f"[Seed] >>> Set seed: {seed}")
    return seed






import numpy as np
import torch
import sklearn.preprocessing as preprocessing # LabelEncoder, Binarizer, StandardScaler, MinMaxScaler, FunctionTransformer
from sklearn.utils._encode import _unique


__all__ = ['X2X', 'Label2Index', 'Binarizer', 'StandardScaler', 'MinMaxScaler']


class X2X(preprocessing.FunctionTransformer):
    """Identity transformation."""

    def partial_fit(self, x) -> None:
        ...

    def transform(self, x):
        return x


class Label2Index(preprocessing.LabelEncoder):
    """Convert labels to indices."""

    def _flatten(self, y):
        return np.ravel(np.asarray(y))

    def partial_fit(self, y):
        y = self._flatten(y)
        if not hasattr(self, 'classes_'):
            self.fit(y)
        else:
            self.classes_ = _unique(np.concatenate([self.classes_, y]))

    def transform(self, y):
        y = self._flatten(y)
        return super().transform(y)


class Binarizer(preprocessing.Binarizer):

    def partial_fit(self, x): ...


class StandardScaler(preprocessing.StandardScaler): ...
class MinMaxScaler(preprocessing.MinMaxScaler): ...


if __name__ == "__main__":

    transformer = Label2Index()
    transformer.partial_fit(torch.tensor([[1], [2], [3]]))
    print(transformer.transform(torch.tensor([[1], [2], [3]])).reshape(-1, 1))




from typing import Union, Dict

import numpy as np

from ..datasets import BaseSet, RecDataSet


__all__ = ['Postprocessor', 'ModeError']


class Postprocessor(BaseSet):

    def __init__(self, datapipe: Union[RecDataSet, 'Postprocessor']) -> None:
        super().__init__()
        self.source = datapipe
        self.fields = self.source.fields

    def train(self):
        super().train()
        self.source.train()
        return self

    def valid(self):
        super().valid()
        self.source.valid()
        return self

    def test(self):
        super().test()
        self.source.test()
        return self

    def at_least_2d(self, array: np.array):
        return array[:, None] if array.ndim == 1 else array

    @property
    def datasize(self):
        return self.source.datasize

    def __len__(self):
        return len(self.source)

class ModeError(Exception): ...






from .base import ImplicitRecSet
from ..fields import SparseField
from ..tags import USER, ITEM, ID
from ...dict2obj import Config


__all__ = ['MovieLens1M_m2']


class MovieLens1M_m2(ImplicitRecSet):
    """ MovieLens1M: (user, item, rating, timestamp)
    |    Dataset     | #Users | #Items | #Interactions | #Train  | #Test  | Density |
    | :------------: | :----: | :----: | :-----------: | :-----: | :----: | :-----: |
    | MovieLens1M_m2 | 6,022  | 3,043  |    895,699    | 796,244 | 99,455 | 0.04888 |
    See [here](https://github.com/openbenchmark/BARS/tree/master/candidate_matching/datasets) for details.

    Attritbutes:
    ---

    _cfg: Config
        - sparse: SparseField
            UserID + ItemID
        - dense: DenseField
            Timestamp
        - target: SparseField
            Rating
    open_kw: Config
        - mode: 'rt'
        - delimiter: '\\t'
        - skip_lines: 0

    """

    URL = "https://zenodo.org/record/7297855/files/MovieLens1M_m2.zip"

    _cfg = Config(
        sparse = [
            SparseField(name='UserID', na_value=None, dtype=int, tags=[USER, ID]),
            SparseField(name='ItemID', na_value=None, dtype=int, tags=[ITEM, ID]),
        ]
    )

    _cfg.fields = _cfg.sparse





from .base import ImplicitRecSet
from ..fields import SparseField
from ..tags import USER, ITEM, ID
from ...dict2obj import Config


__all__ = ['Yelp18_m1']


class Yelp18_m1(ImplicitRecSet):
    """ Yelp 2018 dataset.
    |  Dataset  | #Users | #Items | #Interactions |  #Train   |  #Test  | Density |
    |:---------:|:------:|:------:|:-------------:|:---------:|:-------:|:-------:|
    | Yelp18_m1 | 31,668 | 38,048 |   1,561,406   | 1,237,259 | 324,147 | 0.00130 |

    See [here](https://github.com/openbenchmark/BARS/tree/master/candidate_matching/datasets/Yelp#Yelp18_m1) for details.

    Attributes:
    ---

    _cfg: Config
        - sparse: SparseField
            UserID + ItemID
        - target: SparseField
            Rating
    open_kw: Config
        - mode: 'rt'
        - delimiter: ' '
        - skip_lines: 0

    """

    URL = "https://zenodo.org/record/7297855/files/Yelp18_m1.zip"

    _cfg = Config(
        sparse = [
            SparseField(name='UserID', na_value=None, dtype=int, tags=[USER, ID]),
            SparseField(name='ItemID', na_value=None, dtype=int, tags=[ITEM, ID]),
        ]
    )

    _cfg.fields = _cfg.sparse




from .base import ImplicitRecSet
from ..fields import SparseField
from ..tags import USER, ITEM, ID
from ...dict2obj import Config


__all__ = ['Gowalla_m1']


class Gowalla_m1(ImplicitRecSet):
    """ GowallaM1: (user, items).
    |  Dataset   | #Users | #Items | #Interactions | #Train  |  #Test  | Density |
    | :--------: | :----: | :----: | :-----------: | :-----: | :-----: | :-----: |
    | Gowalla_m1 | 29,858 | 40,981 |   1,027,370   | 810,128 | 217,242 | 0.00084 |

    See [here](https://github.com/kuandeng/LightGCN/tree/master/Data/gowalla) for details.

    Attributes:
    ---

    _cfg: Config
        - sparse: SparseField
            UserID + ItemID
        - target: SparseField
            Rating
    open_kw: Config
        - mode: 'rt'
        - delimiter: ' '
        - skip_lines: 0

    """

    URL = "https://zenodo.org/record/7184851/files/Gowalla_m1.zip"

    _cfg = Config(
        sparse = [
            SparseField(name='UserID', na_value=None, dtype=int, tags=[USER, ID]),
            SparseField(name='ItemID', na_value=None, dtype=int, tags=[ITEM, ID]),
        ]
    )

    _cfg.fields = _cfg.sparse

    open_kw = Config(mode='rt', delimiter=' ', skip_lines=0)




from typing import Iterator, Dict

import torchdata.datapipes as dp
import numpy as np
import random
import scipy.sparse as sp
from math import ceil

from .base import Postprocessor, ModeError
from ..fields import SparseField
from ..tags import USER, ITEM, ID
from ...utils import errorLogger, timemeter


__all__ = ['NegativesForTrain', 'NegativesForEval', 'UniformSampler', 'TriSampler']


@dp.functional_datapipe("negatives_for_train_")
class NegativesForTrain(Postprocessor):
    """Sampling negatives for trainpipe."""

    def __init__(
        self, datapipe: Postprocessor, 
        num_negatives: int = 1,
        unseen_only: bool = True
    ) -> None:
        """
        Parameters:
        ---

        datapipe: RecDataSet or Postprocessor
            Yielding dict of np.array.
        num_negatives: int
            Sampling `num_negatives` for every piece of data.
        unseen_only: bool, default True
            - `True`: Sampling negatives only from unseen items (slow).
            - `False`: Sampling negatives from all items (fast).
        """
        super().__init__(datapipe)
        self.num_negatives = num_negatives
        self.unseen_only = unseen_only
        self.User: SparseField = self.fields[USER, ID]
        self.Item: SparseField = self.fields[ITEM, ID]
        self.prepare()

    @timemeter("NegativeForTrain/prepare")
    def prepare(self):
        self.train()
        self.posItems = [set() for _ in range(self.User.count)]
        self.negative_pool = self._sample_from_all(self.datasize * self.num_negatives)

        for chunk in self.source:
            list(map(
                lambda row: self.posItems[row[0].item()].add(row[1].item()),
                zip(chunk[self.User.name], chunk[self.Item.name])
            ))

    def _sample_from_all(self, pool_size: int = 51200):
        allItems = list(range(self.Item.count))
        while 1:
            for item in random.choices(allItems, k=pool_size):
                yield item

    def _sample_from_pool(self, seen):
        negative = next(self.negative_pool)
        while negative in seen:
            negative = next(self.negative_pool)
        return negative

    def sample(self, x):
        seen = self.posItems[x.item()]
        return [self._sample_from_pool(seen) for _ in range(self.num_negatives)]

    def __iter__(self) -> Iterator:
        if self.mode == 'train':
            for chunk in self.source:
                if self.unseen_only:
                    negatives = self.at_least_2d(
                        np.apply_along_axis(self.sample, 1, chunk[self.User.name])
                    )
                else:
                    negatives = np.random.randint(
                        0, self.Item.count, 
                        size=(len(chunk[self.User.name]), self.num_negatives)
                    )
                chunk[self.Item.name] = np.concatenate([chunk[self.Item.name], negatives], axis=1)
                yield chunk
        else:    
            raise ModeError(errorLogger("for training only ..."))


@dp.functional_datapipe("negatives_for_eval_")
class NegativesForEval(NegativesForTrain):
    """Sampling negatives for valid|testpipe."""

    def __init__(self, datapipe: Postprocessor, num_negatives: int = 1) -> None:
        """
        Parameters:
        ---

        datapipe: RecDataSet or Postprocessor
            Yielding dict of np.array.
        num_negatives: int
            `num_negatives` negatives will be sampled for every user in advance, 
            and then they will be yieled following their users. 
            Note that all negatives will be frozen once sampled.
        """
        super().__init__(datapipe, num_negatives)

    @timemeter("NegativeForEval/prepare")
    def prepare(self):
        self.train()
        posItems = [set() for _ in range(self.User.count)]
        allItems = set(range(self.Item.count))
        self.negItems = []

        for chunk in self.source:
            list(map(
                lambda row: posItems[row[0].item()].add(row[1].item()),
                zip(chunk[self.User.name], chunk[self.Item.name])
            ))
        for items in posItems:
            unseen = list(allItems - items)
            self.negItems.append(random.sample(unseen, k=self.num_negatives))
        self.negItems = np.array(self.negItems)

    def __iter__(self) -> Iterator:
        if self.mode in ('valid', 'test'):
            for chunk in self.source:
                negatives = self.negItems[np.ravel(chunk[self.User.name])]
                chunk[self.Item.name] = np.concatenate([chunk[self.Item.name], negatives], axis=1)
                yield chunk
        else:
            raise ModeError(errorLogger("for evaluation only ..."))


@dp.functional_datapipe("uniform_sampling_")
class UniformSampler(Postprocessor):
    """Uniformly sampling users and their negatives."""

    def __init__(
        self, datapipe: Postprocessor, 
        num_negatives: int = 1,
        unseen_only: bool = True
    ) -> None:
        """
        Parameters:
        ---

        datapipe: RecDataSet or Postprocessor
            Yielding dict of np.array.
        num_negatives: int
            Sampling `num_negatives` for every piece of data.
        unseen_only: bool
            - `True`: Sampling negatives only from unseen items (slow).
            - `False`: Sampling negatives from all items (fast).
        """
        super().__init__(datapipe)
        self.num_negatives = num_negatives
        self.unseen_only = unseen_only
        self.User: SparseField = self.fields[USER, ID]
        self.Item: SparseField = self.fields[ITEM, ID]
        self.fields = self.fields.groupby(ID)
        self.prepare()

    @timemeter("UniformSampler/prepare")
    def prepare(self):
        self.train()
        self.posItems = [set() for _ in range(self.User.count)]
        self.negative_pool = self._sample_from_all(self.datasize * self.num_negatives)

        for chunk in self.source:
            list(map(
                lambda row: self.posItems[row[0].item()].add(row[1].item()),
                zip(chunk[self.User.name], chunk[self.Item.name])
            ))

        self.posItems = [list(items) for items in self.posItems]

    def _sample_from_all(self, pool_size: int = 51200):
        allItems = list(range(self.Item.count))
        while 1:
            for item in random.choices(allItems, k=pool_size):
                yield item

    def _sample_from_pool(self, seen):
        negative = next(self.negative_pool)
        while negative in seen:
            negative = next(self.negative_pool)
        return negative

    def sample_from_seen(self, user):
        return random.choice(self.posItems[user.item()])

    def sample_from_unseen(self, user):
        seen = self.posItems[user.item()]
        return [self._sample_from_pool(seen) for _ in range(self.num_negatives)]

    def __len__(self):
        return 1

    def __iter__(self):
        if self.mode == 'train':
            users = np.random.randint(0, self.User.count, (self.datasize, 1))
            positives = self.at_least_2d(np.apply_along_axis(self.sample_from_seen, 1, users))
            if self.unseen_only:
                negatives = self.at_least_2d(
                    np.apply_along_axis(self.sample_from_unseen, 1, users)
                )
            else:
                negatives = np.random.randint(
                    0, self.Item.count, 
                    size=(self.datasize, self.num_negatives)
                )
            yield {self.User.name: users, self.Item.name: np.concatenate((positives, negatives), axis=1)}
        else:
            raise ModeError(errorLogger("for training only ..."))


@dp.functional_datapipe("trisample_")
class TriSampler(Postprocessor):
    """Yielding users with all items.
    The items are trinary including -1, 0, 1:
        - `-1`: This item has been used in trainset;
        - `0`: A negative item.
        - `1`: A positive item has not used in trainset.
    """

    def __init__(
        self, datapipe: Postprocessor, batch_size: int
    ) -> None:
        """
        Parameters:
        ---

        datapipe: RecDataSet or Postprocessor
            Yielding dict of np.array.
        batch_size: int
        """
        super().__init__(datapipe)
        self.batch_size = batch_size
        self.User: SparseField = self.fields[USER, ID]
        self.Item: SparseField = self.fields[ITEM, ID]
        self.fields = self.fields.groupby(ID)
        self.prepare()

    @timemeter("TriSampler/prepare")
    def prepare(self):
        users, items, vals = [], [], []

        self.train()
        for chunk in self.source:
            users.append(chunk[self.User.name])
            items.append(chunk[self.Item.name])
            # -1 for seen positives
            vals.append(-np.ones_like(chunk[self.Item.name])) 

        self.test()
        for chunk in self.source:
            users.append(chunk[self.User.name])
            items.append(chunk[self.Item.name])
            # 1 for unseen positives
            vals.append(np.ones_like(chunk[self.Item.name]))
        
        users = np.ravel(np.concatenate(users, axis=0))
        items = np.ravel(np.concatenate(items, axis=0))
        vals = np.ravel(np.concatenate(vals, axis=0))

        self.graph = sp.csr_array(
            (vals, (users, items)), dtype=np.int64
        )


    def __len__(self):
        return ceil(self.User.count / self.batch_size)

    def __iter__(self):
        if self.mode == 'train':
            raise ModeError(errorLogger("for evaluation only ..."))
        else:
            allUsers = np.arange(self.User.count)
            for k in range(len(self)):
                start = k * self.batch_size
                end = (k + 1) * self.batch_size
                users = allUsers[start:end]
                items = self.graph[users].todense()
                yield {self.User.name: users[:, None], self.Item.name: items}





import torch

def _collate_fn(batch):
    return batch[0]

class DataLoader(torch.utils.data.DataLoader):

    def __init__(self, datapipe, num_workers: int = 0, **kwargs) -> None:
        super().__init__(
            dataset=datapipe, num_workers=num_workers, 
            batch_size=1, collate_fn=_collate_fn, **kwargs
        )





from typing import List
import torch
import torch.nn as nn
# from torchrec.modules.deepfm import FactorizationMachine


__all__ = ['FM']



class FM(nn.Module):
    """Tensor: B x [d1 + d2 ...] -> Tensor: B x 1"""

    def forward(
        self,
        inputs,
    ) -> torch.Tensor:
        # flatten each embedding to be [B, N, D] -> [B, N*D], then cat them all on dim=1
        fm_input = inputs
        sum_of_input = torch.sum(fm_input, dim=1, keepdim=True)
        sum_of_square = torch.sum(fm_input * fm_input, dim=1, keepdim=True)
        square_of_sum = sum_of_input * sum_of_input
        cross_term = square_of_sum - sum_of_square
        cross_term = torch.sum(cross_term, dim=1, keepdim=True) * 0.5  # [B, 1]
        return cross_term






from typing import Callable, Iterator, Union

import torchdata.datapipes as dp

from .base import Postprocessor
from ..datasets import RecDataSet


__all__ = ['Batcher', 'Mapper', 'ShardingFilter', 'Prefetcher']


@dp.functional_datapipe("batch_")
class Batcher(Postprocessor):

    def __init__(
        self, datapipe: Union[RecDataSet, Postprocessor],
        batch_size: int, drop_last: bool = False
    ) -> None:
        super().__init__(datapipe)
        self.datapipe = datapipe.batch(batch_size, drop_last)

    def __iter__(self) -> Iterator:
        yield from self.datapipe


@dp.functional_datapipe("map_")
class Mapper(Postprocessor):

    def __init__(
        self, datapipe: Union[RecDataSet, Postprocessor], fn: Callable
    ) -> None:
        super().__init__(datapipe)
        self.datapipe = datapipe.map(fn)
    
    def __iter__(self) -> Iterator:
        yield from self.datapipe


@dp.functional_datapipe("sharding_filter_")
class ShardingFilter(Postprocessor):

    def __init__(self, datapipe: Union[RecDataSet, Postprocessor]) -> None:
        super().__init__(datapipe)
        self.datapipe = datapipe.sharding_filter()

    def __iter__(self) -> Iterator:
        yield from self.datapipe


@dp.functional_datapipe("prefetch_")
class Prefetcher(Postprocessor):

    def __init__(self, datapipe: Union[RecDataSet, Postprocessor], buffer_size: int = 2) -> None:
        super().__init__(datapipe)
        self.datapipe = datapipe.prefetch()

    def __iter__(self) -> Iterator:
        yield from self.datapipe




import torch
import os, argparse, time, yaml
from argparse import ArgumentParser

from .dict2obj import Config
from .utils import (
    mkdirs, timemeter, set_color, set_seed, activate_benchmark, 
    set_logger, errorLogger
)


DATA_DIR = 'data'
SUMMARY_DIR = 'summary'
CHECKPOINT_PATH = "./infos/{description}/{dataset}/{device}"
LOG_PATH = "./logs/{description}/{dataset}/{device}-{id}"
CORE_CHECKPOINT_PATH = "./infos/{description}/core"
CORE_LOG_PATH = "./logs/{description}/core"
TIME = "%m%d%H%M%S"

CONFIG = Config(
    BENCHMARK = True, # activate cudnn.benchmark if True
    SEED = 1, # -1 for random

    # evaluation
    EVAL_FREQ = 5,
    EVAL_VALID = True,
    EVAL_TEST = False,

    # logger
    log2file = True,
    log2console = True,

    # path|file
    SAVED_FILENAME = "model.pt", # the filename of saved model paramters
    CHECKPOINT_FREQ = 1,
    CHECKPOINT_MODULES = ['model', 'optimizer', 'lr_scheduler'],
    CHECKPOINT_FILENAME = "checkpoint.tar",
    SUMMARY_FILENAME = "SUMMARY.md",
    MONITOR_FILENAME = "monitors.pickle",
    MONITOR_BEST_FILENAME = "best.pickle",
    description = "RecSys"
)

CORE_CONFIG = Config(
    MONITOR_BEST_FILENAME = CONFIG.MONITOR_BEST_FILENAME,
    CHECKPOINT_FILENAME = CONFIG.CHECKPOINT_FILENAME,
    EXCLUSIVE = False,
    COMMAND = None,
    ENVS = dict(),
    PARAMS = dict(),
    DEFAULTS = dict(),

    log2file = True,
    log2console = True,
)


class Parser(Config):
    """ArgumentParser wrapper."""

    def __init__(self) -> None:
        super().__init__(**CONFIG)
        self.parse()

    def readme(self, path: str, mode: str = "w") -> None:
        """Add README.md to the path."""
        time_ = time.strftime("%Y-%m-%d-%H:%M:%S")
        file_ = os.path.join(path, "README.md")
        s = "|  {key}  |   {val}    |\n"
        info = "\n## {0} \n\n\n".format(time_)
        info += "|  Attribute   |   Value   |\n"
        info += "| :-------------: | :-----------: |\n"
        for key, val in self.items():
            info += s.format(key=key, val=val)
        with open(file_, mode, encoding="utf8") as fh:
            fh.write(info)

    @timemeter("Parser/load")
    def load(self, args: ArgumentParser):
        """Load config.yaml."""
        if hasattr(args, 'config') and args.config:
            with open(args.config, encoding="UTF-8", mode='r') as f:
                for key, val in yaml.full_load(f).items():
                    if key.upper() in self:
                        self[key.upper()] = val
                    elif key in self:
                        self[key] = val
                    else:
                        errorLogger(f"Unexpected parameter of {key} from {args.config} ...")

    def reset(self):
        self.clear()
        for key, val in CONFIG.items():
            self[key] = val

    @timemeter("Parser/parse")
    def parse(self):

        self.parser = argparse.ArgumentParser()

        self.parser.add_argument("--root", type=str, default=".", help="data")
        self.parser.add_argument("--dataset", type=str, default="RecDataSet", help="useless if no need to automatically select a dataset")
        self.parser.add_argument("--config", type=str, default=None, help="config.yml")

        self.parser.add_argument("--device", default=torch.cuda.current_device() if torch.cuda.is_available() else 'cpu', help="device")

        # model
        self.parser.add_argument("--optimizer", type=str, default="adam", help="Optimizer: adam (default), sgd, ...")
        self.parser.add_argument("--nesterov", action="store_true", default=False, help="nesterov for SGD")
        self.parser.add_argument("-mom", "--momentum", type=float, default=0.9, help="the momentum used for SGD")
        self.parser.add_argument("-beta1", "--beta1", type=float, default=0.9, help="the first beta argument for Adam")
        self.parser.add_argument("-beta2", "--beta2", type=float, default=0.999, help="the second beta argument for Adam")
        self.parser.add_argument("-wd", "--weight-decay", type=float, default=None, help="weight for 'l1|l2|...' regularzation")
        self.parser.add_argument("-lr", "--lr", "--LR", "--learning-rate", type=float, default=None)
        self.parser.add_argument("-b", "--batch-size", type=int, default=None)
        self.parser.add_argument("--epochs", type=int, default=None)

        # eval
        self.parser.add_argument("--eval-valid", action="store_false", default=True, help="evaluate validset")
        self.parser.add_argument("--eval-test", action="store_true", default=False, help="evaluate testset")
        self.parser.add_argument("--eval-freq", type=int, default=CONFIG.EVAL_FREQ, help="the evaluation frequency")

        self.parser.add_argument("--num-workers", type=int, default=0)

        self.parser.add_argument("--seed", type=int, default=CONFIG.SEED, help="calling --seed=-1 for a random seed")
        self.parser.add_argument("--benchmark", action="store_false", default=True, help="cudnn.deterministic == True ?")
        self.parser.add_argument("--verbose", action="store_true", default=False, help="show the progress bar if true")
        self.parser.add_argument("--resume", action="store_true", default=False, help="resume the training from the recent checkpoint")

        self.parser.add_argument("--id", type=str, default=time.strftime(TIME))
        self.parser.add_argument("-m", "--description", type=str, default=CONFIG.description)

    def add_argument(self, *args: str, **kwargs):
        args = (arg.replace('_', '-') for arg in args) # user '-' instead of '_'
        self.parser.add_argument(*args, **kwargs)

    def set_defaults(self, **kwargs):
        self.parser.set_defaults(**kwargs)

    @timemeter("Parser/compile")
    def compile(self):
        """Generate config file according to settings.

        Flows:
        ---

        1. Load (default) settings from parsed args (ArgumentParser).
        2. Load settings from xxx.yaml if --config has been activated. 
        Note that this step might override the settings above.
        3. Generate basic paths:
            - CHECKPOINT_PATH: saving checkpoints
            - LOG_PATH: collecting training infomation

        4. Set Logger, and then you can log information by info|debug|warn|errorLogger ...
        5. Finally, READMD.md will be added under CHECKPOINT_PATH and LOG_PATH both.
        """
        args = self.parser.parse_args()
        for key, val in args._get_kwargs():
            if key.upper() in self:
                self[key.upper()] = val
            else:
                self[key] = val
        self.load(args) # loading config (.yaml) ...

        try:
            self.device = int(self.device)
        except ValueError:
            ...
        
        self['DATA_DIR'] = DATA_DIR
        self['SUMMARY_DIR'] = SUMMARY_DIR
        self['CHECKPOINT_PATH'] = CHECKPOINT_PATH.format(**self)
        self['LOG_PATH'] = LOG_PATH.format(**self)
        mkdirs(
            self.CHECKPOINT_PATH, self.LOG_PATH,
            os.path.join(self.LOG_PATH, self.DATA_DIR),
            os.path.join(self.LOG_PATH, self.SUMMARY_DIR)
        )
        set_color(self.device)
        set_logger(path=self.LOG_PATH, log2file=self.log2file, log2console=self.log2console)

        activate_benchmark(self.BENCHMARK)
        self.SEED = set_seed(self.SEED)

        self.readme(self.CHECKPOINT_PATH) # generate README.md
        self.readme(self.LOG_PATH)


class CoreParser(Config):
    ALL_ENVS = (
        'description', 'root', 'device', 'eval_freq', 'num_workers'
    )

    def __init__(self) -> None:
        super().__init__(**CORE_CONFIG)
        self.parse()

    @timemeter("CoreParser/parse")
    def parse(self):

        self.parser = argparse.ArgumentParser()

        self.parser.add_argument("description", type=str, help="...")
        self.parser.add_argument("config", type=str, help="config.yml")
        self.parser.add_argument("--exclusive", action="store_true", default=False, help="one by one or one for all")

        self.parser.add_argument("--root", type=str, default=None, help="data")
        self.parser.add_argument("--dataset", type=str, default=None, help="useless if no need to automatically select a dataset")
        self.parser.add_argument("--device", type=str, default=None, help="device")

        self.parser.add_argument("--eval-freq", type=int, default=None, help="the evaluation frequency")

        self.parser.add_argument("--num-workers", type=int, default=None)

        self.parser.add_argument("--resume", action="store_true", default=False, help="resume the search from the recent checkpoint")


    def check(self):
        """Check the validity of the given config."""
        template = """
        Please make sure the configuration file follows the template below:

        command: python xxx.py
        envs:
            root: ../../data
            dataset: Gowalla_m1
            device: 0,1,2,3
            eval_freq: 5
            num_workers: 0
        params:
            optimizer: [adam, sgd]
            learning_rate: [1e-3, 1e-2, 1e-1]
            weight_decay: [0, 1e-4, 2e-4, 5e-4]
            batch_size: [128, 256, 512, 1024]
        defaults:
            optimizer: adam
            learning_rate: 1e-3
            weight_decay: 0
            batch_size: 256
            epochs: 100
            seed: 1

        where 'command' is necessary but 'envs', 'params' and 'defaults' are optional ...

        Notes: when calling '--exclusive' for grid search one by one, 
            'defaults' is required for clear comparsions in tensorbaord.
        """
        if self.COMMAND is None:
            errorLogger(template)
        for key in ('root', 'device'):
            if self.ENVS.get(key, None) is None:
                errorLogger(f"No {key} is allocated, calling '--{key}' to specify it")

        if self.ENVS.get('dataset', None) is None:
            self.ENVS['dataset'] = "RecDataSet"
        self.ENVS = Config(self.ENVS)
        self.PARAMS = Config(self.PARAMS)

    @timemeter("Parser/load")
    def load(self, args: ArgumentParser):
        """Load config."""
        with open(args.config, encoding="UTF-8", mode='r') as f:
            config = {key.upper(): vals for key, vals in yaml.full_load(f).items()}
            self.update(**config)
        self.EXCLUSIVE = args.exclusive
        self.resume = args.resume
        for key, val in args._get_kwargs():
            if key in self.ALL_ENVS and val is not None:
                self.ENVS[key] = val

    @timemeter("CoreParser/compile")
    def compile(self):
        """Generate config file according to settings.

        Flows:
        ---

        1. Load settings from xxx.yaml which provides parameters for grid searching.
        2. Load settings of the execution environment from parsed args (ArgumentParser).
        3. Generate basic paths:
            - CHECKPOINT_PATH: subprocess
            - LOG_PATH: subprocess
            - CORE_CHECKPOINT_PATH: saving checkpoints of the rest of params
            - CORE_LOG_PATH: saving best results of each subprocess for comparison

        4. Set Logger, and then you can log information by info|debug|warn|errorLogger ...
        5. Finally, READMD.md will be added under CHECKPOINT_PATH and LOG_PATH both.
        """
        args = self.parser.parse_args()
        self.load(args)
        self.check()

        self['DATA_DIR'] = DATA_DIR
        self['SUMMARY_DIR'] = SUMMARY_DIR
        self['CHECKPOINT_PATH'] = CHECKPOINT_PATH
        self['LOG_PATH'] = LOG_PATH
        self['CORE_CHECKPOINT_PATH'] = CORE_CHECKPOINT_PATH.format(**self.ENVS)
        self['CORE_LOG_PATH'] = CORE_LOG_PATH.format(**self.ENVS)
        mkdirs(self.CORE_CHECKPOINT_PATH, self.CORE_LOG_PATH)
        set_logger(path=self.CORE_LOG_PATH, log2file=self.log2file, log2console=self.log2console)

        self.readme(self.CORE_LOG_PATH)

    def readme(self, path: str, mode: str = "w") -> None:
        """Add README.md to the path."""
        time_ = time.strftime("%Y-%m-%d-%H:%M:%S")
        file_ = os.path.join(path, "README.md")
        s = "|  {key}  |   {val}    |\n"
        info = "\n## {0} \n\n\n".format(time_)
        info += "|  Attribute   |   Value   |\n"
        info += "| :-------------: | :-----------: |\n"
        for key, val in self.ENVS.items():
            info += s.format(key=key, val=val)
        for key, val in self.PARAMS.items():
            info += s.format(key=key, val=val)
        for key, val in self.DEFAULTS.items():
            info += s.format(key=f"{key} (default)", val=val)
        with open(file_, mode, encoding="utf8") as fh:
            fh.write(info)




from typing import Iterator, Dict, Optional, Tuple

import torch, os, random
import numpy as np
import torchdata.datapipes as dp
from torch_geometric.data import Data, HeteroData
from torch_geometric.utils import to_undirected
from freeplot.utils import import_pickle, export_pickle
from math import ceil

from ..tags import SPARSE, USER, ITEM, ID
from ..fields import Fielder, SparseField
from ..utils import collate_dict, download_from_url, extract_archive
from ...utils import timemeter, infoLogger, errorLogger, mkdirs, warnLogger
from ...dict2obj import Config


__all__ = ['BaseSet', 'RecDataSet']


_DEFAULT_PICKLE_FMT = "{0}_from_pickle"
_DEFAULT_TRANSFORM_FILENAME = "transforms.pickle"
_DEFAULT_CHUNK_FMT = "chunk{0}.pickle"


class BaseSet(dp.iter.IterDataPipe):

    def __init__(self) -> None:
        super().__init__()

        self.__mode = 'train'

    @property
    def fields(self):
        return self.__fields

    @fields.setter
    def fields(self, vals) -> Fielder:
        self.__fields = Fielder(vals)
        
    @property
    def mode(self):
        return self.__mode

    def train(self):
        self.__mode = 'train'
        return self

    def valid(self):
        self.__mode = 'valid'
        return self

    def test(self):
        self.__mode = 'test'
        return self

    def __len__(self):
        raise NotImplementedError()

    @timemeter("DataSet/to_graph")
    def to_heterograph(self, *edge_types: Tuple[SparseField, Optional[str], SparseField]) -> HeteroData:
        """Convert datapipe to a heterograph.

        Parameters:
        ---

        *edge_types: (src, edge, dst)
            - src: SparseField
                Source node.
            - edge: Optional[str]
                The name of the edge. `src.name2dst.name` will be specified if `edge` is `None`.
            - dst: SparseField
                Destination node.

        Notes:
        ---

        Warning will be raised if current mode is not 'train' !

        Examples:
        ---

        >>> from freerec.data.datasets import Gowalla_m1
        >>> from freerec.data.tags import USER, ITEM, ID
        >>> basepipe = Gowalla_m1("../data")
        >>> fields = basepipe.fields
        >>> graph = basepipe.to_heterograph(
        ...    (fields[USER, ID], None, fields[ITEM, ID]), 
        ...    (fields[ITEM, ID], None, fields[USER, ID])
        ... )
        >>> graph
        HeteroData(
            UserID={ x=[29858, 0] },
            ItemID={ x=[40981, 0] },
            (UserID, UserID2ItemID, ItemID)={ edge_index=[2, 810128] },
            (ItemID, ItemID2UserID, UserID)={ edge_index=[2, 810128] }
        )
        """
        if self.mode != 'train':
            warnLogger(f"Convert the datapipe for {self.mode} to graph. Ensure this is intentional ...")

        srcs, _, dsts = zip(*edge_types)
        edges = list(map(lambda triplet: triplet[1] if triplet[1] else f"{triplet[0].name}2{triplet[2].name}", edge_types))
        data = {node.name: [] for node in set(srcs + dsts)}
        for chunk in self:
            for node in data:
                data[node].append(np.ravel(chunk[node]))
        for key in data:
            data[key] = torch.tensor(np.concatenate(data[key], axis=0), dtype=torch.long)

        graph = HeteroData()
        for node in srcs:
            graph[node.name].x = torch.empty((node.count, 0), dtype=torch.long)
        for node in dsts:
            if node not in srcs:
                graph[node.name].x = torch.empty((node.count, 0), dtype=torch.long)
        for src, edge, dst in zip(srcs, edges, dsts):
            u, v = data[src.name], data[dst.name]
            graph[src.name, edge, dst.name].edge_index = torch.stack((u, v), dim=0) # 2 x N
        return graph.coalesce()

    def to_bigraph(
        self,
        src: SparseField,
        dst: SparseField,
        edge_type: Optional[str] = None
    ) -> HeteroData:
        """Convert datapipe to a bipartite graph.

        Parameters:
        ---

        src: SparseField
            Source node.
        dst: SparseField
            Destination node.
        edge_type: Optional[str]
            The name of the edge. `src.name2dst.name` will be specified if `edge` is `None`.

        Notes:
        ---

        Warning will be raised if current mode is not 'train' !

        Examples:
        ---

        >>> from freerec.data.datasets import Gowalla_m1
        >>> from freerec.data.tags import USER, ITEM, ID
        >>> basepipe = Gowalla_m1("../data")
        >>> fields = basepipe.fields
        >>> graph = basepipe.to_bigraph(
        ...    (fields[USER, ID], None, fields[ITEM, ID]), 
        ...    (fields[ITEM, ID], None, fields[USER, ID])
        ... )
        >>> graph
        HeteroData(
            UserID={ x=[29858, 0] },
            ItemID={ x=[40981, 0] },
            (UserID, UserID2ItemID, ItemID)={ edge_index=[2, 810128] }
        )
        """
        return self.to_heterograph((src, edge_type, dst))
   
    def to_graph(self, src: SparseField, dst: SparseField) -> Data:
        """Convert datapipe to a homogeneous graph.

        Parameters:
        ---

        src: SparseField
            Source node.
        dst: SparseField
            Destination node.

        Notes:
        ---

        Warning will be raised if current mode is not 'train' !

        Examples:
        ---

        >>> from freerec.data.datasets import Gowalla_m1
        >>> from freerec.data.tags import USER, ITEM, ID
        >>> basepipe = Gowalla_m1("../data")
        >>> fields = basepipe.fields
        >>> graph = basepipe.to_bigraph(
        ...    (fields[USER, ID], None, fields[ITEM, ID]), 
        ...    (fields[ITEM, ID], None, fields[USER, ID])
        ... )
        >>> graph
        Data(edge_index=[2, 1620256], x=[70839, 0], node_type=[70839], edge_type=[810128])
        """
        graph = self.to_heterograph((src, None, dst)).to_homogeneous()
        graph.edge_index = to_undirected(graph.edge_index)
        return graph

    def summary(self):
        infoLogger(str(self))


class RecDataSet(BaseSet):
    """ RecDataSet provides a template for specific datasets.

    Attributes:
    ---

    _cfg: Config[str, Field]
        Includes fields of each column.
    _DEFAULT_CHUNK_SIZE: int, defalut 51200
        Chunk size for saving.

    Notes:
    ---

    All datasets inherit RecDataSet should define the class variable of `_cfg` before instantiation.
    Generally speaking, the dataset will be splited into 
        - trainset
        - validset
        - testset

    Because these three datasets share the same _cfg, compiling any one of them will overwrite it ! 
    """

    _DEFAULT_CHUNK_SIZE = 51200 # chunk size
    URL: str

    def __new__(cls, *args, **kwargs):
        for attr in ('_cfg',):
            if not hasattr(cls, attr):
                errorLogger("_cfg, should be defined before instantiation ...", AttributeError)
        if not hasattr(cls._cfg, 'fields'):
            errorLogger("Fields sorted by column should be given in _cfg ...", AssertionError)
        return super().__new__(cls)

    def __init__(self, root: str, filename: Optional[str] = None, download: bool = True) -> None:
        """
        Parameters:
        ---

        root: str
            The path storing datasets.
        filename: str, optional
            The dirname of the dataset.
            - `None`: Set the classname as the filename.
        
        download: bool
            Download the dataset from a URL.
        """
        super().__init__()
        filename = filename if filename else self.__class__.__name__
        self.path = os.path.join(root, filename)
        self.fields = self._cfg.fields
        self.trainsize: int = 0
        self.validsize: int = 0
        self.testsize: int = 0

        if not os.path.exists(self.path) or not any(True for _ in os.scandir(self.path)):
            if download:
                extract_archive(
                    download_from_url(self.URL, root, overwrite=False),
                    self.path
                )
            else:
                errorLogger(
                    f"No such file of {self.path} or this dir is empty ...",
                    FileNotFoundError
                )
        self.compile()

    @property
    def cfg(self):
        """Return the config of the dataset."""
        return self._cfg

    def check_transforms(self):
        """Check if the transformations exist."""
        file_ = os.path.join(
            self.path,
            _DEFAULT_PICKLE_FMT.format(self.__class__.__name__), 
            _DEFAULT_TRANSFORM_FILENAME
        )
        if os.path.isfile(file_):
            return True
        else:
            mkdirs(os.path.join(
                self.path,
                _DEFAULT_PICKLE_FMT.format(self.__class__.__name__)
            ))
            return False

    def save_transforms(self):
        infoLogger(f"[{self.__class__.__name__}] >>> Save transformers ...")
        state_dict = self.fields.state_dict()
        state_dict['trainsize'] = self.trainsize
        state_dict['validsize'] = self.validsize
        state_dict['testsize'] = self.testsize
        file_ = os.path.join(
            self.path, 
            _DEFAULT_PICKLE_FMT.format(self.__class__.__name__), 
            _DEFAULT_TRANSFORM_FILENAME
        )
        export_pickle(state_dict, file_)

    def load_transforms(self):
        file_ = os.path.join(
            self.path, 
            _DEFAULT_PICKLE_FMT.format(self.__class__.__name__), 
            _DEFAULT_TRANSFORM_FILENAME
        )
        state_dict = import_pickle(file_)
        self.trainsize = state_dict['trainsize']
        self.validsize = state_dict['validsize']
        self.testsize = state_dict['testsize']
        self.fields.load_state_dict(state_dict, strict=False)

    def check_pickle(self):
        """Check if the dataset has been converted into feather format."""
        path = os.path.join(
            self.path,
            _DEFAULT_PICKLE_FMT.format(self.__class__.__name__), 
            self.mode
        )
        if os.path.exists(path):
            return any(True for _ in os.scandir(path))
        else:
            os.makedirs(path)
            return False

    def write_pickle(self, data: Dict, count: int):
        """Save pickle format data."""
        file_ = os.path.join(
            self.path,
            _DEFAULT_PICKLE_FMT.format(self.__class__.__name__),
            self.mode, _DEFAULT_CHUNK_FMT.format(count)
        )
        export_pickle(data, file_)

    def read_pickle(self, file_: str):
        """Load pickle format data."""
        return import_pickle(file_)

    def raw2data(self) -> dp.iter.IterableWrapper:
        errorLogger(
            "raw2data method should be specified ...",
            NotImplementedError
        )

    def row_processer(self, row):
        """Row processer for raw data."""
        return {field.name: field.caster(val) for val, field in zip(row, self.fields)}

    def raw2pickle(self):
        """Convert raw data into pickle format."""
        infoLogger(f"[{self.__class__.__name__}] >>> Convert raw data ({self.mode}) to chunks in pickle format")
        datapipe = self.raw2data()
        count = 0
        for chunk in datapipe.batch(batch_size=self._DEFAULT_CHUNK_SIZE).collate(collate_dict):
            for field in self.fields:
                chunk[field.name] = field.transform(chunk[field.name][:, None]).ravel()[:, None] # N x 1
            self.write_pickle(chunk, count)
            count += 1
        infoLogger(f"[{self.__class__.__name__}] >>> {count} chunks done")

    def pickle2data(self):
        """Read pickle data in chunks."""
        datapipe = dp.iter.FileLister(
            os.path.join(
                self.path,
                _DEFAULT_PICKLE_FMT.format(self.__class__.__name__),
                self.mode
            )
        )
        if self.mode == 'train':
            datapipe.shuffle()
        for file_ in datapipe:
            yield self.read_pickle(file_)


    @timemeter("DataSet/compile")
    def compile(self):
        """Check current dataset and transformations.

        Flows:
        ---

        1. Check whether the transformation has been fitted:
            - `True`: Skip.
            - `False`: Fit the total trainset and the `SPARSE` fields in valid|testset
                to avoid unseen features. This operation will not cause information leakage.
            
        2. Convert each set into pickle format for fast loading.

        """

        def fit_transform(fields):
            datapipe = self.raw2data().batch(batch_size=self._DEFAULT_CHUNK_SIZE).collate(collate_dict)
            datasize = 0
            try:
                for batch in datapipe:
                    for field in fields:
                        field.partial_fit(batch[field.name][:, None])
                    datasize += len(batch[field.name])
            except NameError as e:
                errorLogger(e, NameError)
            return datasize

        if self.check_transforms():
            self.load_transforms()
        else:
            self.train()
            self.trainsize = fit_transform(self.fields)

            # avoid unseen tokens not included in trainset
            self.valid()
            self.validsize = fit_transform(self.fields.groupby(SPARSE))
            self.test()
            self.testsize = fit_transform(self.fields.groupby(SPARSE))

            self.save_transforms()
            
        # raw2pickle
        self.train()
        if not self.check_pickle():
            self.raw2pickle()
        self.valid()
        if not self.check_pickle():
            self.raw2pickle()
        self.test()
        if not self.check_pickle():
            self.raw2pickle()
        self.train()

        self.summary()

    @property
    def datasize(self):
        if self.mode == 'train':
            return self.trainsize
        elif self.mode == 'valid':
            return self.validsize
        else:
            return self.testsize

    def __len__(self):
        return ceil(self.datasize / self._DEFAULT_CHUNK_SIZE)

    def __iter__(self) -> Iterator:
        yield from self.pickle2data()

    def __str__(self) -> str:
        cfg = '\n'.join(map(str, self.cfg.fields))
        return f"[{self.__class__.__name__}] >>> \n" + cfg


class _Row2Pairer(dp.iter.IterDataPipe):

    def __init__(self, datapipe: dp.iter.IterDataPipe) -> None:
        super().__init__()
        self.source = datapipe

    def __iter__(self):
        for row in self.source:
            user = row[0]
            for item in row[1:]:
                if item:
                    yield user, item


class ImplicitRecSet(RecDataSet):
    """Implicit feedback data.
    The data should be collected in the order of users; that is,
    each row represents a user's interacted items.
    """

    open_kw = Config(mode='rt', delimiter=' ', skip_lines=0)

    def file_filter(self, filename: str):
        if self.mode == 'train':
            return 'train' in filename
        else:
            return 'test' in filename

    def raw2data(self) -> dp.iter.IterableWrapper:
        datapipe = dp.iter.FileLister(self.path)
        datapipe = datapipe.filter(filter_fn=self.file_filter)
        datapipe = datapipe.open_files(mode=self.open_kw.mode)
        datapipe = datapipe.parse_csv(delimiter=self.open_kw.delimiter, skip_lines=self.open_kw.skip_lines)
        datapipe = _Row2Pairer(datapipe)
        datapipe = datapipe.map(self.row_processer)
        data = list(datapipe)
        if self.mode == 'train':
            random.shuffle(data)
        datapipe = dp.iter.IterableWrapper(data)
        return datapipe

    def summary(self):
        super().summary()
        from prettytable import PrettyTable
        User, Item = self.fields[USER, ID], self.fields[ITEM, ID]

        table = PrettyTable(['#User', '#Item', '#Interactions', '#Train', '#Test', 'Density'])
        table.add_row([
            User.count, Item.count, self.trainsize + self.test().testsize,
            self.trainsize, self.testsize,
            (self.trainsize + self.testsize) / (User.count * Item.count)
        ])

        infoLogger(table)




import torchdata.datapipes as dp

from .base import RecDataSet
from ..fields import SparseField, DenseField
from ..tags import FEATURE, TARGET, ITEM
from ...dict2obj import Config


__all__ = ['Criteo_x1']


class Criteo_x1(RecDataSet):
    """A Kaggle challenge for Criteo (7:2:1) display advertising.
    |     Dataset      |   Total    |   #Train   | #Validation |   #Test   |
    | :--------------: | :--------: | :--------: | :---------: | :-------: |
    |    Criteo_x1     | 45,840,617 | 33,003,326 |  8,250,124  | 4,587,167 |

    The Criteo dataset is a widely-used benchmark dataset for CTR prediction, 
    which contains about one week of click-through data for display advertising. 
    It has 13 numerical feature fields and 26 categorical feature fields.
    See [here](https://github.com/openbenchmark/BARS/tree/master/ctr_prediction/datasets/Criteo) for details.

    Attributes:
    ---

    _cfg: Config
        - sparse: SparseField
        - dense: DenseField
            Note that all dense features will be normalized by MinMaxScaler in default.
        - target: SparseField
            Label
    open_kw: Config
        - mode: 'rt'
        - delimiter: ','
        - skip_lines: 1
    """

    URL = "https://zenodo.org/record/5700987/files/Criteo_x1.zip"

    _cfg = Config(
        # dataset
        sparse = [SparseField(name=f"C{idx}", na_value='-1', dtype=str, tags=[ITEM, FEATURE]) for idx in range(1, 27)],
        dense = [DenseField(name=f"I{idx}", na_value=0., dtype=float, tags=[ITEM, FEATURE]) for idx in range(1, 14)],
        target = [SparseField(name='Label', na_value=None, dtype=int, transformer='none', tags=TARGET)]
    )
    _cfg.fields = _cfg.target + _cfg.dense + _cfg.sparse

    open_kw = Config(mode='rt', delimiter=',', skip_lines=1)


    def file_filter(self, filename: str):
        return self.mode in filename

    def raw2data(self) -> dp.iter.IterableWrapper:
        datapipe = dp.iter.FileLister(self.path)
        datapipe = datapipe.filter(filter_fn=self.file_filter)
        datapipe = datapipe.open_files(mode=self.open_kw.mode)
        datapipe = datapipe.parse_csv(delimiter=self.open_kw.delimiter, skip_lines=self.open_kw.skip_lines)
        datapipe = datapipe.map(self.row_processer)
        # data = list(datapipe)
        # if self.mode == 'train':
        #     random.shuffle(data)
        # datapipe = dp.iter.IterableWrapper(data)
        return datapipe



import setuptools
import re

with open("README.md", "r") as fh:
  long_description = fh.read()

requires = [
    "torchdata==0.4.1",
    "torchmetrics>=0.9.2",
    "tqdm>=4.64.0",
    'scipy>=1.9.1',
    "freeplot>=0.2.0",
    "PyYAML>=6.0",
    "tensorboard>=2.10.0",
]


def get_property(prop, project):
    result = re.search(r'{}\s*=\s*[\'"]([^\'"]*)[\'"]'.format(prop), open(project + '/__init__.py').read())
    return result.group(1)


setuptools.setup(
  name="freerec",
  version=get_property('__version__', 'freerec'),
  author="MTandHJ",
  author_email="congxueric@gmail.com",
  description="PyTorch library for recommender systems",
  long_description=long_description,
  long_description_content_type="text/markdown",
  license='MIT License',
  url="https://github.com/MTandHJ/freerec",
  packages=setuptools.find_packages(),
  python_requires='>=3.9',
  install_requires=requires,
  classifiers=[
  "Programming Language :: Python :: 3",
  "License :: OSI Approved :: MIT License",
  "Operating System :: OS Independent",
  ],
)




from typing import Iterable, Union

import torch
import torch.nn as nn
import torch.nn.functional as F

from .utils import errorLogger


__all__ = ["BaseCriterion", "BCELoss", "MSELoss", "L1Loss"]


class BaseCriterion(nn.Module):
    """Criterion."""

    def __init__(self, reduction: str = 'mean') -> None:
        super().__init__()
        assert reduction in ('none', 'sum', 'mean'), f"Invalid reduction of {reduction} got ..."
        self.reduction = reduction

    @staticmethod
    def regularize(params: Union[torch.Tensor, Iterable[torch.Tensor]], rtype: str = 'l2'):
        """Add regularization for given parameters.

        Parameters:
        ---

        params: List of parameters for regularization.
        rtype: Some kind of regularization including 'l1'|'l2'.
        """
        params = [params] if isinstance(params, torch.Tensor) else params
        if rtype == 'l1':
            return sum(param.abs().sum() for param in params)
        elif rtype == 'l2':
            return sum(param.pow(2).sum() for param in params) / 2
        else:
            torch.norm()
            errorLogger(f"{rtype} regularization is not supported ...", NotImplementedError)


class BCELoss4Logits(BaseCriterion):
    """Binary Cross Entropy with logits !!!"""

    def forward(self, logits: torch.Tensor, targets: torch.Tensor):
        return F.binary_cross_entropy_with_logits(logits, targets.to(logits.dtype), reduction=self.reduction)
    

class BPRLoss(BaseCriterion):

    def forward(self, pos_scores: torch.Tensor, neg_scores: torch.Tensor):
        """
        Parameters:
        ---

        pos_scores: torch.Tensor
            positive scores
        neg_scores: torch.Tensor
            negative scores
        
        Returns:
        ---

            torch.Tensor
        """
        loss = F.softplus(neg_scores - pos_scores)
        if self.reduction == 'none':
            return loss
        elif self.reduction == 'mean':
            return loss.mean()
        elif self.reduction == 'sum':
            return loss.sum()


class MSELoss(BaseCriterion):
    """Mean Square Loss"""

    def forward(self, inputs: torch.Tensor, targets: torch.Tensor):
        return F.mse_loss(inputs, targets, reduction=self.reduction)


class L1Loss(BaseCriterion):

    def forward(self, inputs: torch.Tensor, targets: torch.Tensor):
        return F.l1_loss(inputs, targets, reduction=self.reduction)





from typing import TypeVar, Callable, Dict, List, Optional

import numpy as np
import os, requests, warnings, hashlib, tqdm

from ..utils import errorLogger, infoLogger


T = TypeVar('T')

def safe_cast(val: T, dest_type: Callable[[T], T], default: T) -> T:
    """Cast the value to the specified type.

    Flows:
    ---

    1. If the `val` is not `None` or `''` -> return `dest_type(val)`;
    2. Else return `dest_type(default)`.

    Raises:
    ---

    ValueError:
        1. `val` does not belong to the type `dest_type`;
        2. `default` does not belong to the type `dest_type`;
    Note that the second case is sometimes useful if you want to make sure that there are no "null" values in your data.
    """
    try:
        if val:
            return dest_type(val)
        else: # fill_na
            if default is None:
                raise ValueError
            return dest_type(default)
    except (ValueError, TypeError):
        errorLogger(
            f"Using '{dest_type.__name__}' to convert '{val}' where the default value is {default} ..." \
            f"This happens when the value (or the default value: {default}) to be cast is not of the type {dest_type.__name__}.",
            ValueError
        )


def collate_dict(batch: List[Dict]):
    elem = batch[0]
    return {key: np.array([d[key] for d in batch]) for key in elem}


def download_from_url(
    url: str, root: str = '.', filename: Optional[str] = None, 
    overwrite: bool = False, retries=5, 
    sha1_hash: Optional[str] = None, verify_ssl: bool = True, 
    log: bool = True
):
    """Download a given URL.

    Codes borrowed from dgl.data.utils

    Parameters
    ---

    url : str
        URL to download.
    root : str, default '.'
        Destination path to store downloaded file.
    filename: str, optional
        Filename of the downloaded file.
    overwrite : bool, optional
        Whether to overwrite the destination file if it already exists.
    sha1_hash : str, optional
        Expected sha1 hash in hexadecimal digits. Will ignore existing file when hash is specified
        but doesn't match.
    retries : integer, default 5
        The number of times to attempt downloading in case of failure or non 200 return codes.
    verify_ssl : bool, default True
        Verify SSL certificates.
    log : bool, default True
        Whether to print the progress for download

    Returns
    ---

    str
        The file path of the downloaded file.
    """
    if filename is None:
        filename = url.split('/')[-1]
        # Empty filenames are invalid
        assert filename, 'Can\'t construct file-name from this URL. ' \
            'Please set the `path` option manually.'
    file_ = os.path.join(root, filename)

    assert retries >= 0, "Number of retries should be at least 0"

    if not verify_ssl:
        warnings.warn(
            'Unverified HTTPS request is being made (verify_ssl=False). '
            'Adding certificate verification is strongly advised.')

    if overwrite or not os.path.exists(file_) or (sha1_hash and not check_sha1(file_, sha1_hash)):
        dirname = os.path.dirname(os.path.abspath(os.path.expanduser(file_)))
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        while retries+1 > 0:
            # Disable pyling too broad Exception
            # pylint: disable=W0703
            try:
                if log:
                    infoLogger('[DataSet] >>> Downloading %s from %s...' % (file_, url))
                r = requests.get(url, stream=True, verify=verify_ssl)
                if r.status_code != 200:
                    errorLogger("Failed downloading url %s" % url, RuntimeError)
                with open(file_, 'wb') as f:
                    for chunk in tqdm.tqdm(r.iter_content(chunk_size=1024), leave=False, desc="վ'ᴗ' ի-"):
                        if chunk:  # filter out keep-alive new chunks
                            f.write(chunk)
                if sha1_hash and not check_sha1(file_, sha1_hash):
                    errorLogger(
                        'File {} is downloaded but the content hash does not match.'
                        ' The repo may be outdated or download may be incomplete. '
                        'If the "repo_url" is overridden, consider switching to '
                        'the default repo.'.format(file_)
                    )
                break
            except Exception as e:
                retries -= 1
                if retries <= 0:
                    raise e
                else:
                    if log:
                        infoLogger("[DataSet] >>> Download failed, retrying, {} attempt{} left"
                              .format(retries, 's' if retries > 1 else ''))

    return file_


def extract_archive(file_, target_dir, overwrite=False):
    """Extract archive file.

    Codes borrowed from dgl/data/utils.py

    Parameters
    ---

    file_ : str
        Absolute path of the archive file.
    target_dir : str
        Target directory of the archive to be uncompressed.
    overwrite : bool, default True
        Whether to overwrite the contents inside the directory.
        By default always overwrites.
    """
    if os.path.exists(target_dir) and not overwrite:
        return
    infoLogger('[DataSet] >>> Extracting file to {}'.format(target_dir))
    if file_.endswith('.tar.gz') or file_.endswith('.tar') or file_.endswith('.tgz'):
        import tarfile
        with tarfile.open(file_, 'r') as archive:
            archive.extractall(path=target_dir)
    elif file_.endswith('.gz'):
        import gzip
        import shutil
        with gzip.open(file_, 'rb') as f_in:
            target_file = os.path.join(target_dir, os.path.basename(file_)[:-3])
            with open(target_file, 'wb') as f_out:
                shutil.copyfileobj(f_in, f_out)
    elif file_.endswith('.zip'):
        import zipfile
        with zipfile.ZipFile(file_, 'r') as archive:
            archive.extractall(path=target_dir)
    else:
        errorLogger('Unrecognized file type: ' + file_)


def check_sha1(filename, sha1_hash):
    """Check whether the sha1 hash of the file content matches the expected hash.

    Codes borrowed from dgl/data/utils.py

    Parameters
    ---

    filename : str
        Path to the file.
    sha1_hash : str
        Expected sha1 hash in hexadecimal digits.

    Returns
    ---

    bool
        Whether the file content matches the expected hash.
    """
    sha1 = hashlib.sha1()
    with open(filename, 'rb') as f:
        while True:
            data = f.read(1048576)
            if not data:
                break
            sha1.update(data)

    return sha1.hexdigest() == sha1_hash






from typing import Optional, Union
import torch
import torch.nn as nn

__all__ = ['RecSysArch']


class RecSysArch(nn.Module):

    def to(
        self, device: Optional[Union[int, torch.device]] = None,
        dtype: Optional[Union[torch.dtype, str]] = None,
        non_blocking: bool = False
    ):
        if device:
            self.device = device
        return super().to(device, dtype, non_blocking)

    def initialize(self):
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.xavier_normal_(m.weight)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0.)
            elif isinstance(m, nn.Embedding):
                nn.init.normal_(m.weight, std=1e-4)
            elif isinstance(m, (nn.BatchNorm1d, nn.BatchNorm2d)):
                nn.init.constant_(m.weight, 1.)
                nn.init.constant_(m.bias, 0.)


    @staticmethod
    def broadcast(*tensors: torch.Tensor):
        """Broadcasts the given tensors according to Broadcasting semantics.
        See [here](https://pytorch.org/docs/stable/generated/torch.broadcast_tensors.html#torch.broadcast_tensors) for details.

        Examples:
        ---

        >>> users = torch.rand(4, 1, 4)
        >>> items = torch.rand(4, 2, 1)
        >>> users, items = RecSysArch.broadcast(users, items)
        >>> users.size()
        torch.Size(4, 2, 4)
        >>> items.size()
        torch.Size(4, 2, 4)
        """
        return torch.broadcast_tensors(*tensors)




from typing import Dict, List

import torch
import torch.nn as nn

from .base import RecSysArch

from ..data.fields import Tokenizer
from ..data.tags import SPARSE, DENSE, FEATURE
from ..layers import FM


__all__ = ['DeepFM']

class DeepFM(RecSysArch):

    def __init__(self, tokenizer: Tokenizer) -> None:
        super().__init__()

        self.tokenizer = tokenizer
        sparse_dim = self.tokenizer.calculate_dimension(FEATURE, SPARSE)
        dense_dim = self.tokenizer.calculate_dimension(FEATURE, DENSE)
        self.linear = nn.Linear(sparse_dim + dense_dim, 1, bias=False)
        self.fm = FM()
        self.dnn = nn.Sequential(
            nn.Linear(sparse_dim + dense_dim, 64),
            # nn.BatchNorm1d(64),
            nn.ReLU(True),
            # nn.Dropout(0.)
            nn.Linear(64, 64),
            # nn.BatchNorm1d(64),
            nn.ReLU(True),
            # nn.Dropout(0.)
            nn.Linear(64, 64),
            # nn.BatchNorm1d(64),
            nn.ReLU(True),
            # nn.Dropout(0.)
            nn.Linear(64, 1)
        )

        self.initialize()

    
    def flatten_cat(self, inputs: List[torch.Tensor]) -> torch.Tensor:
        """ List[torch.Tensor: B x k x d] -> torch.Tensor: B x (k1 x d1 + k2 x d2 + ...)"""
        return torch.cat([input_.flatten(1) for input_ in inputs], dim=1)

    def forward(self, users: Dict[str, torch.Tensor], items: Dict[str, torch.Tensor]) -> torch.Tensor:
        sparse: List[torch.Tensor] = self.tokenizer.look_up(items, FEATURE, SPARSE)
        dense: List[torch.Tensor] = self.tokenizer.look_up(items, FEATURE, DENSE)

        outs_linear = self.linear(self.flatten_cat(sparse + dense))
        outs_fm = self.fm(self.flatten_cat(sparse))
        outs_dnn = self.dnn(self.flatten_cat(sparse + dense))
        outs = outs_linear + outs_fm + outs_dnn
        if self.training:
            return outs
        else:
            return outs.sigmoid()



